# ANOVA of balanced factorial designs
#
# Efficient version for large data matrix, no design matrix involved
# Estimation of missing values by anova model
#
# General conventions:
#
# a - data matrix, names(dim(a)) contains the dimension names that
#   might be specified in a formula
# formula - a model formula without response variable
#   (which is given automatically by the matrix entries)
#   eg ~(V+G+S/A)^2 - G:A%in%S
# anv - the bfaov object
# inters - interactions (effects) as derived from the formula
#   a list of vectors of dimension indices, eg list(1,2,3,c(1,2),c(1,3))
#   effects 1,2,3 variable and interaction effects between 1,2 and 1,3 
# comblist - a list of combinations of interactions for nested effects
#   a list of vectors of inters indices, eg list(c(3,5))
#   combine inters effects 3 and c(1,3) in a common nested effect
#
# L Wernisch, 1/6/2001


anv.setup <- function(a,inters,comblist) {
  #anv$inters are always sorted
  y.. <- sum.na(a)
  mu <- y../prod(dim(a))
  mu.s <- (y..*y..)/prod(dim(a))
  #cat('anv.setup-done\n')
  list("inters"=lapply(inters,sort),comblist=comblist,
      "mu"=mu,"mu.s"=mu.s,
      "E"=list(),"SS"=vector(),"df"=vector())
}

anv.finish <- function(anv,a,formula=NULL,drop.array=F) {
  # We don't want to keep matrix a and therefore
  # extract all neccessary information here and store it in anv

  d.last <- length(dim(a))
  yijkn <- apply(a,1:d.last,sum)
  SS.T <- sum.na(a*a) - anv$mu.s
  # last SS is SS.E = SS.T - rest
  anv$SS <- c(anv$SS,SS.T - sum(anv$SS))
  df.T <- prod(dim(a)) - 1
  red.df <- ifelse(is.null(attr(a,"reduce.df")),
      0, attr(a,"reduce.df") )
  anv$df <- c(anv$df,df.T - sum(anv$df) - red.df)
  MSE <- anv$SS/anv$df
  s <- length(MSE)
  F <- MSE/MSE[s]
  list("dims"=dim(a),"inters"=anv$inters,comblist=anv$comblist,
       coordnames=names(dim(a)),
       formula=formula,anova.formula=NULL,
       data=(if (!drop.array) a else NULL),
       "mu"=anv$mu,"mu.s"=anv$mu.s,
       "E"=anv$E,"SS"=anv$SS,"df"=anv$df,
       "MSE"=MSE,"F"=MSE/MSE[length(MSE)],"p"=1-pf(F,anv$df,anv$df[s]) )
} 

expand.array <- function(E,dims,inter.full,inter) {
  # Blow up a lower dimensional array
  # dims - the underlying array dims
  # inter.full - superset interactions
  # inter - subset interactions
  if (!(setequal(intersect(inter,inter.full),inter))) {
    cat(inter.full,"\n",inter,"\n")
    stop("expand.array warning: interactions not nested\n")
  }
  v.diff <- setdiff(inter.full,inter)
  # append copies into missing dimensions (as final dimensions)
  O <- outer(E,array(1,c(dims[v.diff])))
  v.perm <- c(inter,v.diff)
  # move new dimensions to proper place
  # reorder according to dimnumber in inter.full
  aperm(O,match(inter.full,v.perm))
}

anv.inters <- function(anv,a,k) {
  # Calculate E, SS etc for interaction number k

  inters <- anv$inters
  #
  #cat(length(inters))
  #cat(k)
  #cat('anv.inters\n')
  d <- inters[[k]]
  dims <- dim(a)
  y.bar <- apply(a,d,sum.na)/prod(dims[-d]) # or apply mean
  E.contained <- array(anv$mu,dims[d])
  #df.contained <- 1
  for (i in seq(along=inters)) {
    if (is.subset(inters[[i]],d) && !setequal(inters[[i]],d)) {
      E.contained <-
          E.contained + expand.array(anv$E[[i]],dims,d,inters[[i]])
      #df.contained <- df.contained + anv$df[i]
    }
  }
  E <- y.bar - E.contained
  S <- prod(dims[-d])*sum(E^2)
  anv$E <- c(anv$E,list(E))
  anv$SS <- c(anv$SS,S)
  anv$df <- c(anv$df,prod(dims[d] - 1))
  #anv$df <- c(anv$df,prod(dims[d]) - df.contained)
  #anv$E[[k]] <- E
  #anv$SS[[k]] <- S
  #anv$df[[k]] <- prod(dims[d] - 1)
  anv
}

matrix.from.inters <- function(inters,n=NULL) {
  if (is.null(n))
    n <- max(sapply(inters,max))
  inter.vec <- function(inter,n) {
    v <- rep(0,n); v[inter] <- 1; v
  }
  t(sapply(inters,inter.vec,n))
}

dataframe.from.array <- function(a) {
  exper <- sapply(dim(a),function(d) as.factor(1:d))
  grid <- do.call("expand.grid",exper)
  names(grid) <- names(dim(a))
  data.frame(y=as.vector(a), grid)
}

inters.from.matrix <- function(m,names) {
  # m - the factor matrix of a formula
  # names - a vector of names

  if (missing(names))
    inds <- 1:ncol(m)
  else
    inds <- match(colnames(m),names)
  inters <- list()
  for (i in 1:nrow(m)) {
    #cat(m[i,])
    inter <- inds[which(m[i,] != 0)]
    inters <- c(inters,list(inter))
  }
  inters
}

rewrite.list.rev <- function(l,o) {
  rev.o <- NULL
  rev.o[o] <- 1:length(o)
  lapply(l, function(i) rev.o[i])
}

topsort <- function(inters,comblist) {
  # Sort inters topologically by "subset"
  # and rewrite comblist accordingly

  ones <- sapply(inters,length)
  o <- order(ones)
  list(inters=inters[o],
       comblist=rewrite.list.rev(comblist,o))
}

align.inters <- function(inters,inters.source) {
  # Aligning inters with inters.source and return order

  index.inters.set <- function(v,m) which(sapply(m,setequal,v))
  o <- sapply(inters,index.inters.set,inters.source)
  # some inters vecs missing or double in inters.comb results in "list"
  # not "numeric":
  if (mode(o) == "list") {
    stop("Some effects of formula not found in model: MODIFY FORMULA OR MODEL!")
  }
  o
}

close.formula <- function(formula,names) {
  #cat('close.start\n')
  t <- terms(formula)
  inters <- inters.from.matrix(t(attr(t,"factors")),names)
  #cat('close.inters\n')
  #cat('close.formula.end\n')
  close.inters(inters)
}

close.inters <- function(inters) {
  # A heuristic to guess nested effects:
  # calculate the closure of inters with respect to "subset"
  # ie add all missing subsets whenever some subsets 
  # need to be added as subsets of some interaction i.
  # They together with i form a new combination of nested effects
  # and are stored in comblist
  #for (i in 1:length(inters)){
  #  cat(inters[[i]])
  #}
  #cat('1:\n')
  n <- max(sapply(inters,max))
  #cat(n)
  #cat('2:\n')
  all.subs <- all.subsets(n)
  #cat('3:\n')
  in.inters.set <- function(v,s) {
    any(sapply(s,setequal,v))
  }
  #cat('4:\n')
  len.inters <- length(inters)
  #cat('5:\n')
  comblist <- as.list(1:len.inters)
  #cat('6:\n')
  l.remove <- NULL
  #cat('7:\n')
  for (i in 1:len.inters) {
    v <- inters[[i]]
    #cat(v)
    v.sub.all <- all.subs[sapply(all.subs,is.subset,v)]
    v.sub.need <- v.sub.all[!sapply(v.sub.all,in.inters.set,inters)]
    if (length(v.sub.need) > 0) {
      start <- length(inters)+1
      inters <- c(inters,v.sub.need)
      end <- length(inters)
      l.remove <- c(l.remove,i)
      comblist <- c(comblist,list(c(start:end,i)))
    }
  }
  if (!is.null(l.remove)) comblist <- comblist[-l.remove]
  #cat('8:\n')
  inters <- lapply(inters,sort)
  #cat('9:\n')
  topsort(inters,comblist)
}

#/**
# \name{bfaov}
# \alias{bfaov}
# \alias{set.formula}
# \title{Efficient ANOVA of balanced factorial designs}
# \description{
#   Efficient ANOVA for balanced factorial data matrix, where each
#   dimension corresponds to a main effect. 
# }
# \usage{
#   bfaov(formula,a)
#   bfaov(a)
#   set.formula(anv,formula)
# }
# \arguments{
#   \item{formula}{a model formula without response variable
#     (the matrix entries themselves), eg ~(V+S/A)\^2 + G, all variety
#     V, sample S, and array A (nested in S) main effects and
#     their interaction effects, plus gene G main effect only}
#   \item{a}{a balanced factorial matrix with effects as dimension, as
#     obtained for example by \code{\link{rg2log.array}}. The formula
#     might be attached to \code{a} by \code{\link{set.formula}} in
#     which case it can be omitted. If no formula is given (neither
#     explicitely nor attached) all interactions are assumed.}
#   \item{anv}{a bfaov object as generated by \code{bfaov}}
# }
# \value{
#   bfaov object, can be analysed by \code{\link{anova.bfaov},
#   coefficients are obtained by \code{\link{coef.bfaov}}, for more
#   analysis functions see below}
# }
# \details{
#   Most ANOVA algorithms operate on the \eqn{X'X}{X'X} matrix where
#   \eqn{X}{X} is the design matrix, the size of which is determined
#   by the levels of effects and their interactions. Since gene
#   effects and their interactions can have thousands of levels, this
#   approach is not feasible computationally for microarray
#   experiments. In case of balanced factorial designs there are
#   simple explicit solutions to the normal equations derived from the
#   above matrix which are implemented here. Missing data might be
#   interpolated by \code{\link{interpolate.bfaov}}
# }
# \examples{
#   data(trcs)
#   RG <- rg.remove.quantile(trcs.RG,0.1)
#   a <- rg2log.array(RG)
#   av <- bfaov(~(V+G+S/A)^2,a)
#   anova(av)
#   av <- set.formula(av,~(V+S/A)^2 + G + V:G)
#   anova(av)
# }
# \seealso{
#   \code{\link{rg2log.array}},\code{\link{anova.bfaov}},
#   \code{\link{coef.bfaov}},\code{\link{interpolate.bfaov}}
# }
# \keyword{models}
#*/

bfaov <- function(x,...) {
  if (is.null(class(x)))
    class(x) <- data.class(x)
  # P. Sykacek 2005 -> turn off warnings (they are an effect of allowing
  # depending on the data type two different bfaov calls that require
  # different numbers of parameters and don't match one call
  o.curr <- options()
  options(warn=-1)
  #cat('bfaov\n\n')
  # ret <- NextMethod("bfaov",x,...)
  ret <- UseMethod("bfaov")
  options(o.curr)
  ret
}

bfaov.formula <- function(formula,a) {
  # Anova analysis according to formula
  # formula - see beginning of this file
  # The resulting bfaov object av can be inspected by
  #   anova(av)
  # Note: anova() allows to specify a formula as well,
  # this means the formula here should be as comprehensive as
  # possible, partial effects can then be studied using an anova formula
  #cat('bfaov.formula\n\n')
  #cat(names(dim(a)))
  if (is.null(names(dim(a))))
    stop("array needs dimension names names(dim(a))")
  #t <- terms(formula)
  #inters <- inters.from.matrix(t(attr(t,"factors")),names(dim(a)))
  #cat('close.formula\n\n')
  closed <- close.formula(formula,names(dim(a)))
  #cat('bfaov.array\n\n')
  res <- bfaov.array(a,closed,formula=formula)
  res$call <- match.call()
  res
}

bfaov.array <- function(a,closed=NULL,inters=NULL,
     formula=NULL,drop.array=F) {
  # Anova of balanced factorial design given by an array
  # Interactions are added so that inters is closed under "subset"
  # inters - see beginning of this file
  # if missing all possible interactions are taken
  #cat('bfaov.array\n\n')
  if (missing(closed) && is.null(inters))
    closed <- close.inters(all.subsets(length(dim(a))))
  if (!is.null(inters)) {
    closed$inters <- inters
    closed$comblist <- as.list(1:length(inters))
  }
  #cat('anv.setup\n\n')
  anv <- anv.setup(a,closed$inters,closed$comblist)
  for (i in 1:(length(closed$inters))) {
    anv <- anv.inters(anv,a,i)
  }
  #cat('anv.finish\n\n')
  res <- anv.finish(anv,a,formula=formula,drop.array)
  #cat('bfaov.array-done\n\n')
  class(res) <- "bfaov"
  res
}

set.formula <- function(anv,formula) {
  anv$anova.formula <- formula
  anv
}

#/**
# @name
# \title{ANOVA table of balanced factorial designs}
# \description{
#   Prints ANOVA table of bfaov object \code{\link{bfaov}}
# }
# @usage
# \arguments{
#   \item{anv}{a bfaov object, see \code{\link{bfaov}}}
#   \item{formula}{a model formula, see \code{\link{bfaov}}}
#   \item{sigma.sq}{estimate of \eqn{\sigma^2}{sigma^2} for the
#     information criterion measures of goodness of fit}
#   \item{ptype}{print type, if set to "tex", a tex style table is
#     produced}
#   \item{file}{output file, if "" output goes to console}
# }
# \value{
#   if ptype is "none", a list of
#   \item{df}{residual degrees of freedom}
#   \item{SS}{residual sum of squares}
#   \item{MS}{mean residual sum of squares, SS/df, an estimate of
#     \eqn{\sigma^2}{sigma^2} of the ANOVA model}
#   else no return value
# }
# \examples{
#   data(trcs)  
#   RG <- rg.remove.quantile(trcs.RG,0.1)
#   a <- rg2log.array(RG)
#   av <- bfaov(~(V+S/A)^2 + G,a)
#   anova(av)
# }
# \seealso{
#   \code{\link{bfaov}}
# }
# \keyword{print}
#*/

anova.bfaov <- function(anv,formula=NULL,sigma.sq=0,
    ptype=c("console","tex","tex.p","none"),file="") {
  # Print Anova diagnostics
  # anv - an object of class bfaov
  # formula - see beginning of this file
  # sigma.sq - estimate of true error variance sigma^2 for IC calculations
  # ptype - output type

  ptype <- match.arg(ptype)
  if (is.null(formula))
    formula <- anv$anova.formula
  cl <- get.comblist(anv,formula)

  anova.bfaov.comblist(anv,cl,sigma.sq,ptype,file=file)
}

get.comblist <- function(anv,formula=NULL,inters=NULL) {
  if (is.null(formula))
    formula <- anv$anova.formula
  if (is.null(formula) && is.null(inters))
    return(anv$comblist)
  if (!is.null(inters))
    closed <- close.inters(inters)
  else if (!is.null(formula))
    closed <- close.formula(formula,anv$coordnames)

  if (is.null(closed$inters))
      stop("effects missing from analysis")
  o <- align.inters(closed$inters,anv$inters)
  lapply(closed$comblist,function(i) o[i])
}

anova.from.comblist <- function(anv,comblist=NULL) {
  if (is.null(comblist))
    comblist <- anv$comblist
  SS <- NULL
  df <- NULL
  MS <- NULL
  inters <- NULL
  for (i in seq(along=comblist)) {
    comb <- comblist[[i]]
    SS.i <- sum(anv$SS[comb])
    df.i <- sum(anv$df[comb])
    inter <- NULL
    for (j in seq(along=comb))
      inter <- union(inter,anv$inters[[comb[j]]])
    inters <- c(inters,list(inter))
    SS <- c(SS,SS.i)
    df <- c(df,df.i)
  }
  total.inters <- 1:(length(anv$inters) + 1)
  error.inters <- which(!(total.inters %in% unlist(comblist)))
  comblist <- c(comblist,list(error.inters))

  df.e <- sum(anv$df[error.inters])
  SS.e <- sum(anv$SS[error.inters])

  df <- c(df,df.e)
  SS <- c(SS,SS.e)
  # for additional error interactions more something like
  # inters <- c(inters,list(union(inters(error.inters))))

  SS.T <- sum.na(anv$data*anv$data) - anv$mu.s
  SS.ratio <- sum(SS)/SS.T
  if (SS.ratio < 0.99 || SS.ratio > 1.01)
    warning("SS total not equal to sum(SS) in anova.from.comblist")

  list(df=df,SS=SS,MS=SS/df,inters=inters,SS.T=SS.T,length=length(df))
}

model.IC <- function(n,p,SSE,SST,sigma.sq,file=file) {
  # Print various information criteria

  facat <- function(...) cat(...,file=file,append=T)
  facat("p: ",p, " n: ",n," R^2: ",1-SSE/SST)
  facat("  R^2(adj): ",1-(SSE/(n-p))/(SST/(n-1)),"\n")
  facat("Sp: ",(n+p)*SSE/(n-p))
  facat(" Cp: ",SSE + 2*sigma.sq*p)
  facat(" AIC: ",log(SSE/n) + 2*p/n)
  facat(" Schwarz IC: ",log(SSE/n) + log(n)*p/n)
  facat(" SHQ IC: ",log(SSE/n) + 2*log(log(n))*p/n)
  facat("\n")
}

tex.intro <- "Include in latex header:\n\n\\usepackage{dcolumn}\n\\newcolumntype{.}{D{.}{.}{-1}}\n\n"

tex.p.preamble <- "\\begin{tabular}{lr....}\neffects & df & \\multicolumn{1}{c}{SS} & \\multicolumn{1}{c}{MS} & \n  \\multicolumn{1}{c}{F} & \\multicolumn{1}{c}{$p$}  \\\\\n\\hline\n"

tex.preamble <- "\\begin{tabular}{lr..}\neffects & df & \\multicolumn{1}{c}{SS} & \\multicolumn{1}{c}{MS} \\\\\n\\hline\n"

tex.conclusion <- "\\end{tabular}\n\n"

mformat <- function(x) {
  if (is.numeric(x) && !is.na(x)) {
    if (x == 0)
      "  < 1.0e-15"
    else
      formatC(x,digits=5,width=11)
  }
  else
    formatC(x,width=10)
}

anova.bfaov.comblist <- function(anv,comblist,sigma.sq=0,ptype,
    file="") {
  # Print anova according to combinations in comblist

  total.inters <- 1:(length(anv$inters) + 1)
  error.inters <- which(!(total.inters %in% unlist(comblist)))
  comblist <- c(comblist,list(error.inters))

  df.e <- sum(anv$df[error.inters])
  SS.e <- sum(anv$SS[error.inters])
  SS.T <- sum.na(anv$data*anv$data) - anv$mu.s
  MSE.e <- SS.e/df.e
  inters <- anv$inters

  if (ptype == "none")
    return(list(df=df.e,SS=SS.e,MS=MSE.e))

  switch (ptype,
    console = cat("effects\t",mformat("df"),mformat("SS"),
        mformat("MS"),mformat("F"),mformat("p"),"\n",sep="",file=file),
    tex = if (file == "") { cat(tex.intro); cat(tex.preamble) }
        else cat(tex.preamble,file=file),
    tex.p = if (file == "") { cat(tex.intro); cat(tex.p.preamble) }
        else cat(tex.p.preamble,file=file)
  )
  for (i in 1:(length(comblist)-1))
    anova.bfaov.line(anv,comblist[[i]],MSE.e,df.e,ptype,
        file=file)
  switch (ptype,
    console = {
        cat("resids  ",mformat(deparse(df.e)),mformat(SS.e),
          mformat(MSE.e),"\n\n",sep="",file=file,append=T );
        model.IC(n = prod(anv$dims), p = prod(anv$dims) - 1 - df.e,
          SSE = SS.e, SST = SS.T, sigma.sq = sigma.sq,file=file )
        },
    tex = cat("\\hline\nresiduals    & ",mformat(deparse(df.e)),"& ",
        mformat(SS.e),"& ",mformat(MSE.e)," \n",tex.conclusion,sep="",
        file=file,append=T ),
    tex.p = cat("\\hline\nresiduals    & ",mformat(deparse(df.e)),"& ",
        mformat(SS.e),"& ",mformat(MSE.e)," \n",tex.conclusion,sep="",
        file=file,append=T ),
    none =
  )
  if (sigma.sq == 0) sigma.sq <- MSE.e
}  

anova.bfaov.line <- function(anv,comb,MSE.e,df.e,ptype,
    file="") {
  df <- sum(anv$df[comb])
  SS <- sum(anv$SS[comb])
  MSE <- SS/df
  inters <- anv$inters
  F <- MSE/MSE.e
  p <- 1 - pf(F,df,df.e)
  if (length(comb) > 1)
    for (i in 1:(length(comb)-1)) {
      cat(anv$coordnames[inters[[comb[i]]]],",",sep="",file=file,append=T)
      #if (i %% 2 == 0) cat("\n")
    }
  switch (ptype,
    console = cat(anv$coordnames[inters[[comb[length(comb)]]]],"\t",
        mformat(deparse(df)),mformat(SS),mformat(MSE),
        mformat(F),mformat(p),"\n",sep="",file=file,append=T ),
    tex = cat(anv$coordnames[inters[[comb[length(comb)]]]],"\t"," & ",
        mformat(deparse(df))," & ",mformat(SS)," & ",mformat(MSE),
        " \\\\\n",sep="",file=file,append=T ),
    tex.p = cat(anv$coordnames[inters[[comb[length(comb)]]]],"\t"," & ",
        mformat(deparse(df))," & ",mformat(SS)," & ",mformat(MSE)," & ",
        mformat(F)," & ",mformat(p)," \\\\\n",sep="",file=file,append=T )
  )
}

#/**
# \name{coef.bfaov}
# \title{Coefficients and residuals of bfaov ANOVA}
# \alias{coef.bfaov}
# \alias{residuals.bfaov}
# \alias{plot.residuals}
# \description{
#    Coefficients and residuals of bfaov object from \code{\link{bfaov}}
# }
# \usage{
#   coef.bfaov(anv,formula=NULL)
#   residuals.bfaov(anv,formula=NULL)
#   plot.residuals(anv,formula=NULL)
# }
# \arguments{
#   \item{anv}{a bfaov object, see \code{\link{bfaov}}}
#   \item{formula}{a model formula, see \code{\link{bfaov}}}
# }
# \value{
#   a matrix with the same dimension as the balanced factorial data
#   matrix (see \code{\link{bfaov}})
# }
# \examples{
#   data(trcs)
#   E <- coef(trcs.av)
#   r <- residuals(trcs.av)
#   a <- trcs.av$data
#   all(r == a - E) # should be TRUE
# }
# \seealso{
#   \code{\link{bfaov}}
# }
# \keyword{models}
#*/

coef.bfaov <- function(anv,formula=NULL) {
  # Returns estimated coefficients (fitted values) of anova model

  formula <- if (!is.null(formula)) formula
    else if (!is.null(anv$anova.formula)) anv$anova.formula
    else NULL
  inters <- if (is.null(formula)) anv$inters
    else close.formula(formula,anv$coordnames)$inters  
  o <- align.inters(inters,anv$inters)
  E <- anv$E
  d <- anv$dims
  inter.full <- 1:length(d)
  Etot <- array(anv$mu,d)
  for (i in seq(along=inters)) {
    Etot <- Etot + expand.array(E[[o[i]]],d,inter.full,inters[[i]])
  }
  Etot
}

residuals.bfaov <- function(anv,formula=NULL) {
  if (is.null(anv$data))
    cat("cannot get array from baof object, probably dropped\n")
    return(0)
  E <- coef(anv,formula)
  anv$data - E
}

plot.residuals <- function(anv,formula=NULL) {
  E <- coef(anv,formula)
  r <- a-E
  plot(E,r,pch=".")
}

# VARIANCE diagnostics and improvement

get.loess <- function(E,r,select=15000,
    ...,span=0.4,prop.exact=1/10) {
  #loess.control(statistics="approximate",trace.hat="approximate")
  if (select > 0 && select < length(E)) {
    q <- prop.exact*select/length(E)
    ind <- which(E < quantile(E,q))
    ind <- union(ind,which(E > quantile(E,1-q)))
    ind.s <- sample(seq(along=E)[-ind],select*(1-2*prop.exact))
    ind <- union(ind,ind.s)
  } else
    ind <- seq(along=E)
  loe <- loess(r[ind]^2 ~ E[ind],deg=1,span=span)
  ind.p <- sample(seq(along=E),min(10000,length(E)))
  plot(E[ind.p],r[ind.p]^2,ylim=c(0,1.1*max(loe$fitted)),pch=".",
      main="",xlab="",ylab="")
  z <- seq(min(E),max(E),length=200)
  pl <- predict.loess(loe,z)
  loess.limit <- 0.05*median(pl)
  pl[pl <= loess.limit] <- loess.limit
  lines(z,pl,col="red",lwd=2)
  cat("fitted range ",range(loe$fitted),"\n")
  list(loe=loe,E=E[ind.p],r=r[ind.p]^2,
      ylim=c(0,1.1*max(loe$fitted)),z=z,pl=pl)
}

plot.loess <- function(gloe,xlab="",ylab="",...) {
  plot(gloe$E,gloe$r,ylim=gloe$ylim,pch=".",main="",
      xlab=xlab,ylab=ylab,...)
  loess.limit <- 0.05*median(gloe$pl)
  gloe$pl[gloe$pl <= loess.limit] <- loess.limit
  lines(gloe$z,gloe$pl,col="red",lwd=2)
}

plot.myloess <- function(loe,frac=1.1,...) {
  yl <- c(0,frac*max(loe$fitted))
  plot(E,r^2,ylim=yl,pch=".")
  z <- seq(min(E),max(E),length=200)
  pl <- predict.loess(loe,z)
  loess.limit <- 0.05*median(pl)
  pl[pl <= loess.limit] <- loess.limit
  lines(z,pl,col="red",lwd=2)
}

box.cox <- function(a,lambda) {
  geom.mean <- exp(mean(log(a)))
  if (lambda != 0)
    (a^lambda - 1)/(lambda * geom.mean^(lambda-1))
  else
    geom.mean*log(a)
}

test.box.cox <- function(a,rg=seq(-1,3,by=0.5),formula,p=0.05) {
  e <- NULL
  for (lambda in rg) {
    av <- bfaov(formula,box.cox(a,lambda))
    res <- anova(av,ptype="none")
    e <- c(e,res$MS)
    df <- res$df
  }
  plot(rg,e,ty="l")
  SSE <- min(e)
  y <- SSE*(1+qf(1-p,1,df)/df)
  lines(range(rg),c(y,y))
  rbind(rg,e)
}

test.loess <- function(a,formula,select=5000) {
  av <- bfaov(formula,a)
  cat("MSE ",anova(av,ptype="none")$MS,"\n")
  E <- coef(av)
  r <- a-E
  get.loess(E,r,select)
}    


# P-VALUEs

#/**
# \name{diff.vg}
# \title{Significance of variety-gene effects from ANOVA analysis}
# \alias{diff.vg}
# \alias{sd.vg}
# \alias{limit.vg}
# \alias{hist.vg}
# \description{
#   Functions to estimate the significance of over- and
#   under-expression from the residual MSE error of the ANOVA model
#   \code{\link{bfaov}}.
# }
# \usage{
#   diff.vg(a,normalize=T)
#   sd.vg(anv,formula=NULL)
#   limit.vg(anv,p.value=0.05,formula=NULL)
#   hist.vg(anv,p.value=0.05,formula=NULL,a=NULL,normalize=T)
# }
# \arguments{
#   \item{a}{a balanced factorial data array, see
#     \code{\link{rg2log.array}} }
#   \item{anv}{a bfaov object, see \code{\link{bfaov}}}
#   \item{p.value}{p-value for significance estimates}
#   \item{formula}{the formula to use for evaluation,
#     if NULL, the anova object formula is used}
#   \item{normalize}{if TRUE, normalization for fold-change}
# }
# \value{
#   \item{diff.vg}{returns the vector of VG effects}
#   \item{sd.vg}{returns the standard deviation derived from the
#     residual MSE of the ANOVA model, which is approximately
#     \eqn{\sqrt{2 MSE/d}}{sqrt(2*MSE/d)}, where \eqn{d}{d} is the
#     number of measurements of a gene on one channel }
#   \item{limit.vg}{returns the limits for over-/under-expression given
#     a particular p-value}
#   \item{hist.vg}{plots the limits from \code{limit.vg} on a
#     histogram of VG effects from \code{diff.vg}}
# }
# \details{
#   The variety-gene effect VG is the difference of the average level
#   of a gene G between the R and G channel corrected for the overall
#   difference of averages of R and G. The residual MSE of the ANOVA
#   model is used to estimate the standard deviation of an expression
#   level. Assuming normal distribution of the residual error p-values
#   for over- and under-expression can be derived.
# }
# \examples{
#   data(trcs)
#   plot(diff.vg(trcs.av$data))
#   hist.vg(trcs.av)
#   limit.vg(trcs.av)
#   sd.vg(trcs.av)
# }
# \seealso{
#   \code{\link{bfaov}}
# }
# \keyword{models}
#*/

diff.vg <- function(a,normalize=T) {
  vdim <- match("V",names(dim(a)))
  gdim <- match("G",names(dim(a)))
  mean.V <- apply(a,vdim,mean)
  mean.VG <- apply(a,c(vdim,gdim),mean)
  # first is R, second G: R/G
  if (normalize) 
    mean.VG[1,] - mean.V[1] - mean.VG[2,] + mean.V[2]
  else
    mean.VG[1,] - mean.VG[2,]
}

df.vg <- function(a) {
  # degrees of freedom for t-distribution
  dims <- dim(a)
  prod(dims[-match(c("V","G"),names(dims))])
}

sd.vg <- function(anv,formula=NULL) {
  # var of estimate should be
  # 2*(1-1/dim[Gdim])*sigma.sq / prod(dim[-c(Vdim,Gdim)]
  a <- anv$data
  dims <- dim(a)
  err <- anova(anv,formula,ptype="none")
  sqrt(as.numeric(2*(1-1/dims[match("G",names(dims))])*err$MS/
        df.vg(a) * 2 ))
}

hist.vg <- function(anv,p.value=0.05,formula=NULL,a=NULL,normalize=T) {
  if (is.null(a) && !is.null(anv$data))
    a <- anv$data
  else
    stop("need array as parameter or in bfaov object")
  hist.vg.v(anv,diff.vg(a,normalize=normalize),
      p.value=p.value,formula=formula)
}

hist.vg.v <- function(anv,vg,p.value=0.05,formula=NULL) {
  q <- limit.vg(anv,p.value=p.value,formula=formula)
  sd <- sd.vg(anv,formula)
  df <- df.vg(anv$data)
  n <- length(vg)
  #q <- abs(qt(p.value/n,df)*sd)
  #q <- abs(qnorm(p.value/n,0,sd))
  hist(vg,breaks=round(n/30),xlim=range(c(vg,q,-q)),
      xlab="",ylab="",main="")
  lines(c(-q,-q),c(0,300),col="red",lwd=2)
  lines(c(q,q),c(0,300),col="red",lwd=2)
  q
}

limit.vg <- function(anv,p.value=0.05,formula=NULL) {
  sd <- sd.vg(anv,formula)
  df <- df.vg(anv$data)
  n <- dim(anv$data)["G"]
  #abs(qt(p.value/n,df)*sd)
  abs(qnorm(p.value/n,0,sd))
}

pval.vg <- function(anv,over=T,formula=NULL,a=NULL,normalize=T) {
  if (is.null(a) && !is.null(anv$data))
    a <- anv$data
  else
    stop("need array as parameter or in bfaov object")
  sd <- sd.vg(anv,formula)
  df <- df.vg(anv$data)
  if (over) 
    #pt(diff.vg(a)/sd,df-1,lower.tail=F)
    pnorm(diff.vg(a,normalize=normalize)/sd,lower.tail=F)
  else
    #pt(diff.vg(a)/sd*sqrt(df),df-1)
    pnorm(diff.vg(a,normalize=normalize)/sd)
}

pval.vg.boots <- function(b,a,over=T,normalize=T) {
  sdb <- if (!is.null(attr(b,"stats.only")))
      b[,2]
    else
      apply(b,1,sd)
  df <- df.vg(a)
  if (over) 
    #pt(diff.vg(a)/sdb,df-1,lower.tail=F)
    pnorm(diff.vg(a,normalize=normalize)/sdb,lower.tail=F)
  else
    #pt(diff.vg(a)/sdb,df-1)
    pnorm(diff.vg(a,normalize=normalize)/sdb)

    #pnorm(0,diff.vg(a),sdb)
    #pnorm(0,diff.vg(a),sdb,lower.tail=F)
}
 



boots.vg <- function(anv,n=50,formula=NULL,
    gloe=NULL,a=NULL,fast=T,stats.only=F) {
  if (is.null(a) && !is.null(anv$data))
    a <- anv$data
  else
    stop("need array as parameter or in bfaov object")
  vdim <- match("V",names(dim(a)))
  gdim <- match("G",names(dim(a)))

  E <- coef(anv,formula)
  r <- a - E
  if (is.null(gloe)) {
    cat("loess fit to residuals, be patient ...\n")
    gloe <- get.loess(E,r)
  }

  pl <- predict.loess(gloe$loe,as.vector(E))
  loess.limit <- 0.05*median(pl)
  pl[pl <= loess.limit] <- loess.limit
  r.smooth <- sqrt(pl)
  r.scaled <- r/r.smooth
  if (!fast) {
    cat("bootstrapping",n,"rounds, be patient\n")
    cat("[consider fast option boots.vg(...,fast=T)] ...\n")
    b <- NULL
    for (i in 1:n) {
      cat (i," ")
      mean.V <- apply(a,vdim,mean)
      mean.VG <- apply(a,c(vdim,gdim),mean)
      b <- cbind( b, mean.VG[1,] - mean.V[1] - mean.VG[2,] + mean.V[2] )
      a <- E + r.smooth*sample(r.scaled,repl=T)
    }
  } else {
    dr <- prod(dim(a)[-c(gdim,vdim)])  # dim rest
    dg <- dim(a)[gdim]  # dim gene
    dv <- dim(a)[vdim]  # dim V
    rest <- seq(along=dim(a))[-c(gdim,vdim)]
    pa <- aperm(a,c(rest,gdim,vdim))
    pE <- aperm(E,c(rest,gdim,vdim))
    r.smooth <- array(r.smooth,dim(a))
    pr.smooth <- aperm(r.smooth,c(rest,gdim,vdim))
    r.scaled <- array(r.scaled,dim(a))
    pr.scaled <- aperm(r.scaled,c(rest,gdim,vdim))
    r <- array(r,dim(a))
    pr.orig <- aperm(r,c(rest,gdim,vdim))

    d.vg <- diff.vg(a)
    # for stats: b[,1] mean, b[,2] sd,
    # b[,3] count over, b[,4] count under
    bdim <- if (stats.only) 4 else n
    b <- matrix( .C("vg_boots",
      as.integer(n),
      as.integer(dr),as.integer(dg),as.integer(dv),
      as.double(pE),
      as.double(d.vg),
      as.double(pr.smooth),as.double(pr.scaled),
      #as.double(rep(1,length(pr.orig))),as.double(pr.orig),
      as.double(rep(0,length(pr.orig))),
      #as.double(pr.orig),
      as.integer(stats.only),
      b=as.double(matrix(0,dg,bdim))
    )$b,dg,bdim)
  }
  if (stats.only) {
    attr(b,"stats.only") <- T
    b[,1] <- b[,1]/n  # mu 
    b[,2] <- sqrt(b[,2]/(n-1) - n*b[,1]^2/(n-1)) # sd
    # should follow a t-distr with df.vg(a)-1 degs of freedom:
    b[,3] <- b[,3]/n  # p over
    b[,4] <- 1-b[,3]  # p under
  }
  cat ("\n")
  list(b=b,gloe=gloe)
}


#/**
# \name{over.under}
# \alias{over.under}
# \alias{over.under.level}
# \title{Analysis of over- and under-expression from ANOVA}
# \description{
#   Calculates p-values for over- and under-expression from ANOVA model.
#   Bootstrapping from residuals can be used to get more
#   reliable results than using residual error estimates.
#   To calculate over- or under-expressed genes use
#   \code{\link{data.over.under}} }
# \usage{
#   over.under(anv,rounds=50,
#       formula=NULL,gloe=NULL,a=NULL,drop.b=F,fast=T,
#       stats.only=T,RG=NULL,normalize=T,span=0.4)
#   over.under.level(anv,rml,formulas,
#       levels.seq,rounds=50,
#       gloe.l=NULL,a=NULL,drop.b=F,
#       stats.only=T,RG=NULL,normalize=T,span=0.4)
# }
# @synopsis
# \arguments{
#   \item{anv}{result of ANOVA analysis, see \code{\link{bfaov}}}
#   \item{pvalue}{siginificance level}
#   \item{rounds}{if nonzero, p-values are also determined by
#     bootstrapping residuals with \code{rounds} resamplings}
#   \item{formula}{a model formula, see \code{\link{bfaov}}}
#   \item{formulas}{a list of a model formulas, describing
#     the residuals for hierarchical bootstrap,
#     see \code{\link{bfaov}}}
#   \item{levels.seq}{a list of main effects, the levels for hierarchical
#     bootstrap}
#   \item{drop.b}{drop bootstrap results (which can become very large
#     depending on \emph{rounds})}
#   \item{rml}{\code{rml} data as return by \code{\link{reml.bfaov}}}
#   \item{gloe}{a loess object from a smooth regression to the
#     residuals (over estimates), for example, from an earlier call to
#     \code{boots.vg}, see also
#     \code{\link{loess}} }
#   \item{gloe.l}{a list of loess objects as in \code{gloe} for hierarchical
#     bootstrap}
#   \item{fast}{if set to T, the compiled bootstrap is called which is
#     \emph{much} faster then the R version, but might not be
#     available on every platform }
#   \item{a}{a balanced factorial data array, see
#     \code{\link{rg2log.array}} }
#   \item{RG}{RG microarray data structure, see \code{\link{read.rg}}}
#   \item{normalize}{}
#   \item{span}{span for loess smoother}
#   \item{stats.only}{if TRUE, only mean and standard deviation of
#     VG or G effect is collected for each bootstrap replicate}
# }
# \details{
#   \code{over.under.level} performs a hierarchical bootstrap as
#   specified in \code{formulas} and \code{levels.seq}.
#   See \code{\link{sd.vg}} for details on p-values for significant
#   differential expression. Bootstrapping residuals can be used to
#   get more reliable p-values }
# }
# \value{
#   List containing:
#   \item{p.over, p.under}{p-values of over-, under-expressed genes}
#   \item{p.over.b, p.under.b}{p-values of over-, under-expressed genes
#     from bootstrap }
#   \item{rounds}{the input parameter}
#   \item{b}{a matrix containing the bootstrap results as columns}
#   \item{gloe}{a loess object from a smooth regression to the
#     residuals for the bootstrap}
#   \item{gloe.l}{a list of loess objects as in \code{gloe} for hierarchical
#     bootstrap}
#   \item{anv}{the input bfaov object}
#   \item{RG}{the input RG object}
# }
# \examples{
#   data(trcs)
#   ou <- over.under(trcs.av,rounds=0)
#   ou$p.over
#
#   RG <- rg.remove.quantile(trcs.RG,0.1)
#   RG <- rg.scale.normalize(RG)
#   a <- rg2log.array(RG,log.ratio=T)
#   av <- bfaov(~G + S/A + G:S + G:S:A,a)
#
#   rml <- reml.bfaov(av,c("S","A"),anova.only=T)
#   print.reml(rml)
#
#   formulas <- list(~G + S + A,
#       ~G + S + A + G:S,~G + S + A + G:S + G:S:A)
#   levels.seq <- c("S","A")
#   ou.b <- over.under.level(av,rml,rounds=500,
#       formulas=formulas,levels.seq=levels.seq,
#       RG=RG,span=0.4,stats.only=F)
#
#   # normality of coefficients obtained by bootstrap:
#   x <- ou.b$b[50,]; qqnorm(x); qqline(x);
#   ks.test(x,"pnorm",mean(x),sd(x)); shapiro.test(x)
#
#   # plot loess fit for residuals:
#   opar <- par(mfrow=c(3,1))
#   for (i in 1:3) plot.loess(ou.b$gloe.l[[i]])
#   par(opar)
# }
# \seealso{
#   \code{\link{data.over.under}}
# }
# \keyword{models}
#*/

over.under <- function(anv,rounds=50,
    formula=NULL,gloe=NULL,a=NULL,drop.b=F,fast=T,
    stats.only=T,RG=NULL,normalize=T,span=0.4) {
  single <- (!is.null(anv)
        && length(grep("V",names(dim(anv$data)))) == 0)
  if (is.null(a) && !is.null(anv$data))
    a <- anv$data
  else
    stop("need array as parameter or in bfaov object")
  if (single) {
    p.over <- pval.g(anv,over=T)
    p.under <- pval.g(anv,over=F)
  } else {
    p.over <- pval.vg(anv,over=T,normalize=normalize)
    p.under <- pval.vg(anv,over=F,normalize=normalize)
  }
  #over <- which(length(p.over)*p.over < pvalue)
  #under <- which(length(p.under)*p.under < pvalue)
  #over <- which(pvalue.fdr(p.over) < pvalue)
  #under <- which(pvalue.fdr(p.under) < pvalue)
  if (rounds <= 0)
    return(list(p.over=p.over,p.under=p.under,anv=anv,RG=RG))

  if (single) {
    bvg <- boots.g(anv,rounds,gloe=gloe,fast=fast,
        stats.only=stats.only,span=span)
    p.over.b <- pval.g.boots(bvg$b,a,over=T)
    p.under.b <- pval.g.boots(bvg$b,a,over=F)
  } else {
    bvg <- boots.vg(anv,rounds,gloe=gloe,fast=fast,
        stats.only=stats.only,span=span)
    p.over.b <- pval.vg.boots(bvg$b,a,over=T,normalize=normalize)
    p.under.b <- pval.vg.boots(bvg$b,a,over=F,normalize=normalize)
  }
  #over.b <- which(length(p.over.b)*p.over.b < pvalue)
  #under.b <- which(length(p.under.b)*p.under.b < pvalue)
  #over.b <- which(pvalue.fdr(p.over.b) < pvalue)
  #under.b <- which(pvalue.fdr(p.under.b) < pvalue)
  list(p.over=p.over,p.under=p.under,
    rounds=rounds,
    p.over.b=p.over.b,p.under.b=p.under.b,
    b=bvg$b,gloe=bvg$gloe,
    anv=anv,RG=RG)
}

over.under.level <- function(anv,rml,formulas,
    levels.seq,rounds=50,
    gloe.l=NULL,a=NULL,drop.b=F,contr.var=NULL,
    stats.only=T,RG=NULL,normalize=T,span=0.4) {
  E.base <- NULL
  if (is.null(a) && !is.null(anv$data))
    a <- anv$data
  if (is.null(E.base)) {
    if (is.null(RG)) {
      stop("either RG or E.base not NULL");
    } else {
      a.i <- rg2log.array(RG,log.intensity=T)
      av.i <- bfaov(~G ,a.i)
      E.base <- coef(av.i)
    }
  }
  else
    stop("need array as parameter or in bfaov object")
  if (!is.null(contr.var)) {
    p.over <- pval.level.contr(anv,rml,contr.var=contr.var,over=T)
    p.under <- pval.level.contr(anv,rml,contr.var=contr.var,over=F)
  } else {
    p.over <- pval.level(anv,rml,over=T)
    p.under <- pval.level(anv,rml,over=F)
  }

  if (rounds <= 0)
    return(list(p.over=p.over,p.under=p.under,anv=anv,RG=RG,rml=rml))

  blg <- boots.level.g(anv,formulas,levels.seq,
      rounds, E.base=E.base,
      gloe.l=gloe.l, stats.only=stats.only,span=span)
  p.over.b <- pval.g.boots(blg$b,a,over=T)
  p.under.b <- pval.g.boots(blg$b,a,over=F)

  list(p.over=p.over,p.under=p.under,
    rounds=rounds,
    p.over.b=p.over.b,p.under.b=p.under.b,
    b=blg$b,gloe.l=blg$gloe.l,
    anv=anv,RG=RG)
}

#/**
# \name{data.over.under}
# \alias{data.over.under}
# \alias{write.over.under}
# \title{Analysis of over- and under-expressed genes}
# \description{
#   Calculates over- or under-expressed genes use from
#   \code{\link{over.under}} structure }
# \usage{
#   data.over.under(ou,genes=NULL,pvalue=0.05,pvalue.u=NULL,
#       sample.values=F,
#       method=c("fdr","bonferroni","none","fdr.nostep"),normalize=T)
#   write.over.under(list.dfr,prefix="",ptype=c("none","tex"),
#       digits=3,genes.orfs=NULL)
# }
# @synopsis
# \arguments{
#   \item{ou}{result from \code{\link{over.under}}
#       or \code{\link{over.under}} }
#   \item{genes}{a vector of gene names}
#   \item{pvalue}{p-value limit for over expression}
#   \item{pvalue.u}{p-value limit for under expression, set
#       set to \code{pvalue} if not specified}
#   \item{method}{method for multiple testing adjustment; if
#     "fdr" Benjamini and Hochberg's step-down method for the False
#     Discovery Rate (see \code{\link{pvalue.fdr}})is used usually
#     resulting in more genes with significant differential
#     expression; "bonferroni" adjusts by the more conservative
#     Bonferroni method }
#   \item{prefix}{file prefix for output files}
#   \item{list.dfr}{result of call to \code{data.over.under}}
#   \item{sample.values}{if TRUE, \code{ou} p-values from bootstrap
#     samples are used instead of normal approximations}
#   \item{normalize}{normalization for fold-change}
#   \item{ptype}{print type}
#   \item{digits}{number of digits}
#   \item{genes.orfs}{a list of two vectors and a string:
#     \code{$orfs} is a list of orf names, such as "Rv0957",
#     \code{$genes} is a list of gene names, such as "purH",
#     and \code{$prefix} is a prefix "Rv"; each orf name is
#     replaced by its gene name if it contains the prefix}
# }
# \details{
#   Prints the result of a call to \code{\link{over.under}} or
#   \code{\link{over.under.level}}.
#   See \code{\link{sd.vg}} for details on p-values for significant
#   differential expression. Bootstrapping residuals can be used to
#   get more reliable p-values
# }
# \value{
#   \code{data.over.under} returns a list of dataframes containing:
#   \item{over, under}{data.frames for over-, under-expressed genes}
#   \item{over.b, under.b}{data.frames for over-, under-expressed genes
#     from bootstrapping residuals }
#   \item{list.dfr}{the result of a call to \code{data.over.under}}
#   \item{prefix}{a path to a directory, prefix to the 4 output files}
# }
# \examples{
#   data(trcs)
#   ou <- over.under(trcs.av,rounds=0)
#   ldf <- data.over.under(ou,genes=trcs.RG$genes,pvalue=0.01)
#   ldf$over
#   # write 4 files in current directory:
#   write.over.under(ldf)
# }
# \seealso{
#   \code{\link{bfaov}},\code{\link{pvalue.fdr}},
#   \code{\link{over.under}}, code{\link{over.under.level}}
# }
# \keyword{models}
#*/

data.over.under <- function(ou,genes=NULL,pvalue=0.05,
    pvalue.u=NULL,sample.values=F,contr.var=NULL,
    method=c("fdr","bonferroni","none","fdr.nostep"),normalize=T ) {
  # required: ou$RG, ou$p.over, ou$p.under
  single <- (!is.null(ou$anv)
        && length(grep("V",names(dim(ou$anv$data)))) == 0)
  a <- ou$anv$data
  if (is.null(a))
    if (is.null(ou$RG))
      stop("need RG")
    else
      a <- rg2log.array(ou$RG)
  if (is.null(genes))
    if (!is.null(ou$RG))
      genes <- ou$RG$genes
    else
      stop("need either genes or RG")
  if (is.null(pvalue.u))
    pvalue.u <- pvalue
  #al <- rg.log.matrix(RG)
  if (single) {
    if (is.null(contr.var))
        fold <- 2^diff.g(a)
    else
        fold <- 2^diff.g.contr(a,contr.var)
  } else
    fold <- 2^diff.vg(a,normalize=normalize)
  
  method <- match.arg(method)
  if (sample.values) {
    p.over.b <- ou$b[,3]
    p.under.b <- ou$b[,4]
  } else {
    p.over.b <- ou$p.over.b
    p.under.b <- ou$p.under.b
  }
  list(
    over = data.ou(genes=genes,pval=ou$p.over,fold=fold,
        niveau=pvalue, method=method ),
    under = data.ou(genes=genes,pval=ou$p.under,fold=fold,
        niveau=pvalue.u, method=method ),
    over.b =
      if (!is.null(ou$p.over.b)) {
        data.ou(genes=genes,pval=p.over.b,fold=fold,
        niveau=pvalue,method=method )
      } else NULL,
    under.b =
      if (!is.null(ou$p.under.b)) {
        data.ou(genes=genes,pval=p.under.b,fold=fold,
        niveau=pvalue.u, method=method )
      } else NULL,
    pvalue=pvalue
  )
}

#/**
# @name
# \title{Benjamini-Hochberg FDR adjustment for multiple testing p-values}
# \description{
#   Applies the FDR adjustment to a vector of p-values.
# }
# @usage
# @synopsis
# \arguments{
#   \item{pvalue}{vector of p-values from mutiple tests that need to be
#     adjusted}
#   \item{step}{if TRUE, monotonicity of p-values is enforced}
# }
# \details{
#   Suppose out of \eqn{R}{R} rejected hypotheses (supposedly
#   significant cases) \eqn{V}{V} are erroneously rejected (not really
#   significant). The false discovery rate is
#   \eqn{Q=V/R}{Q=V/R}. Controlling the FDR at a level of \eqn{p}{p}
#   means that \eqn{p(Q>0)E(Q|Q>0) <= p}{p(Q>0) E(Q|Q>0) <= p}. The
#   test statistics are assumed to be independent.
# }
# \value{
#   a vector of adjusted p-values
# }
# \examples{
#   data(trcs)
#   RG <- rg.remove.quantile(trcs.RG,0.1,level=1)
#   ou <- t.statistic(RG)
#   p.fdr <- pvalue.fdr(ou$p.over)
#   RG$genes[p.fdr < 0.05]
# }
# \references{
#   Benjamini, Y. and Hochberg, Y. (1995), \emph{Controlling the false
#   discovery rate: a practical and powerful approach to multiple
#   testing}, Journal of the Royal Statistical Society, Series B. 57:
#   289--300
# }
# \seealso{
#   \code{\link{t.statistic}},
#   \code{\link{data.over.under}}}
# }
# \keyword{htest}
#*/

pvalue.fdr <- function(pvalue,step=T) {
	o <- order(pvalue)
	n <- length(pvalue)
  pvf <- pvalue[o]*n/(1:n)
  if (step) {
    # Yes, it's step down, see Benjamini Hochberg
    # J Royal Stat Soc 57: p 293
    for (i in (n-1):1)
      pvf[i] <- min(pvf[i],pvf[i+1])
  }
  pv.fdr <- rep(0,n)
  pv.fdr[o] <- pvf
  pv.fdr
}

data.ou <- function(genes,pval,fold,niveau,
    method=c("fdr","bonferroni","none","fdr.nostep")) {
  method <- match.arg(method)
  if (method == "bonferroni")
    pv <-  pmin(length(pval)*pval,1)
  else if (method == "fdr")
    pv <- pmin(pvalue.fdr(pval,step=T),1)
  else if (method == "fdr.nostep")
    pv <- pmin(pvalue.fdr(pval,step=F),1)
  else
    pv <- pval
  inds <- which(pv <= niveau)
  if (length(inds) == 0) {
    warning("no pvalues left on this level\n Increase pvalue or change method!") 
    dfr <- data.frame(number=vector(),name=vector(),
      pvalue=vector(),foldchange=vector(),index=vector())
  } else {
    pvalnew <- pv[inds]; genes <- genes[inds]; fold <- fold[inds]
    if (method == "fdr.nostep")
      o <- order(pval[inds],-fold)
    else
      o <- order(pvalnew,-fold)
    dfr <- data.frame(number=1:length(o),name=genes[o],
        pvalue=pvalnew[o],foldchange=fold[o],index=inds[o] )
    if (method != "none")
      names(dfr)[pmatch("pvalue",names(dfr))] <- method
  }
  dfr
}

tex.ou.preamble.a <- "\\begin{tabular}{rl..}\nnumber & name & \\multicolumn{1}{c}{"
tex.ou.preamble.b <- "} & \\multicolumn{1}{c}{fold-change} \\\\\n\\hline\n"

write.single.table <- function(list.dfr,prefix,ptype=c("none","tex"),
    list.ou,file.ou,digits=3,genes.orfs=NULL) {
  if (is.null(list.dfr[[list.ou]]))
    return(0)
  ptype <- match.arg(ptype)
  if (ptype == "none") {
    sep <- "\t"; eol <- "\n"; ext <- ".dat"; is.cols <- T
  } else if (ptype == "tex") {
    sep <- " & "; eol <- " \\\\\n"; ext <- ".tex"; is.cols <- F
  }
  file <- paste(prefix,file.ou,ext,sep="")
  if (ptype == "tex") {
     meth <- names(list.dfr$over)[3]
     tex.ou.preamble <- 
        paste(tex.ou.preamble.a,meth,tex.ou.preamble.b,sep="")
     cat(tex.ou.preamble,file=file)
  }
  else
     cat("",file=file)
  old.opts <- options(digits=digits)
  if (!is.null(genes.orfs))
    list.dfr[[list.ou]]$name <- alternative.names(
        list.dfr[[list.ou]]$name,genes.orfs,ptype=ptype)
  write.table(list.dfr[[list.ou]][-c(match("index",
          names(list.dfr[[list.ou]])))],,col.name=is.cols,
      file=file,sep=sep,eol=eol,quote=F,row.names=F,append=T )
  if (ptype == "tex") cat(tex.conclusion,file=file,append=T)
  options(old.opts)
}

write.over.under <- function(list.dfr,prefix="",ptype=c("none","tex"),
    digits=3,genes.orfs=NULL) {
  ptype <- match.arg(ptype)
  write.single.table(list.dfr,prefix,ptype=ptype,"over","over",
      digits=digits,genes.orfs=genes.orfs)
  write.single.table(list.dfr,prefix,ptype=ptype,"under","under",
      digits=digits,genes.orfs=genes.orfs)
  write.single.table(list.dfr,prefix,ptype=ptype,"over.b","over_b",
      digits=digits,genes.orfs=genes.orfs)
  write.single.table(list.dfr,prefix,ptype=ptype,"under.b","under_b",
      digits=digits,genes.orfs=genes.orfs)
}

alternative.names <- function(old.genes,genes.orfs,
    ptype=c("none","tex")) {
  ptype <- match.arg(ptype)
  is.tex <- (ptype == "tex")
  gn <- sapply(strsplit(old.genes,"\\.\\("),"[[",1)
  gn <- gsub("\\.","_",gn)
  g.ind <- seq(along=gn)[-grep(genes.orfs$prefix,gn)]
  g.ind.rv <- grep(genes.orfs$prefix,gn)
  for (i in g.ind) {
    g <- gn[i]
    gstr <- ifelse(is.tex,paste("\\textit{",g,"}",sep=""),g)
    o <- genes.orfs$orfs[which(g == genes.orfs$genes)[1]]
    ostr <- ifelse(is.tex,paste("\\textit{",o,"}",sep=""),o)
    gn[i] <- paste(gstr," (",ostr,")",sep="")
  }
  for (i in g.ind.rv) {
    gn[i] <- ifelse(is.tex,paste("\\textit{",gn[i],"}",sep=""),gn[i])
  }
  gn <- gsub("_","\\.",gn)
  gn
}

# Estimation of MISSING VALUES by ANOVA


#/**
# @name
# \title{Interpolate missing values in balanced factorial data array}
# \description{
#   Interpolate missing values in balanced factorial data array from
#   using ANOVA model from \code{\link{bfaov}} }
# @usage
# @synopsis
# \arguments{
#   \item{anv}{a bfaov object as generated by \code{\link{bfaov}},
#     missing values indicated by NA's}
#   \item{a}{a balanced factorial data matrix, see eg
#     \code{\link{rg2log.array}} }
#   \item{fast}{if T, compiled code is used which is \emph{much} faster
#     than the R code but might not be availabel on every platform}
# }
# \details{
#   From the normal equations a system of equations for missing
#   entries can be derived such that the the residual error of the
#   ANOVA model is minimized.
# }
# \value{
#   a balanced factorial data array with interpolated values in place
#   of NA's
# }
# \examples{
#   data(trcs)
#   RG <- rg.remove.quantile(trcs.RG,0.2,level=3)
#   a <- rg2log.array(RG)
#   which(is.na(a))
#   av <- bfaov(~(V+G+S/A)^2,a)
#   a.new <- interpolate.bfaov(av,a)
# }
# \seealso{
#   \code{\link{bfaov}},
#   \code{\link{rg2log.array}}}
# }
# \keyword{models}
#*/

interpolate.bfaov <- function(anv,a,fast=T) {
  if (fast)
    f.interpolate.bfaov(anv,a)
  else
    s.interpolate.bfaov(anv,a)
}

s.interpolate.bfaov <- function(anv,a) {
  # REQUIRE: anv$inters order compatible with is.subset order!
  ind.na <- which(is.na(a))
  m <- length(ind.na)
  Em <- coef(anv)[ind.na] # ANOVA estimates with NA's 0
  inters <- anv$inters
  dims <- dim(a)
  dlist <- list()
  prod.dim.inters <- sapply(inters, function(d) 1/prod(dims[-d]))
  prod.dims <- 1/prod(dims)
  for (i in seq(along=dims))
    dlist <- c(dlist,list(1:dims[i]))
  na.coords <- do.call("expand.grid",dlist)[ind.na,]
  # conversion of data.frame to array for efficiency (big difference!)
  #na.m <- NULL
  #for (i in 1:nrow(na.coords))
  #  na.m <- cbind(na.m,as.numeric(na.coords[,i]))
  na.m <- as.matrix(na.coords)

	len.inters <- length(inters)
  seq.inters <- seq(along=inters)
  is.sub.inters <- matrix(FALSE,length(inters),length(inters))
  # for efficient subset test:
  for (i in seq.inters)
    for (j in seq.inters)
      is.sub.inters[i,j] <- is.subset(inters[[i]],inters[[j]])
  diag(is.sub.inters) <- FALSE

  estimate.interpol.coef <- function(i,y,e,est.coef) {    
    d <- inters[[i]]
    # if there is a match of indices in the relevant dimensions d:
    if (all(na.m[y,d] == na.m[e,d]))
      c <- prod.dim.inters[i]
    else
	    c <- 0

    # Only works with inters ordered topologically !!
    c <- c - sum(est.coef[is.sub.inters[,i]])
    # empty list unfortunately not in inters list
    c <- c - prod.dims  # mu is always contained
    c
  }

  cat("normal equations number: \n")
  norm.coefs <- matrix(prod.dims,m,m)
  for (e in 1:m) {
    cat(e," ")
    for (y in 1:m) {
      est.coef <- rep(0,len.inters)
      for (i in seq.inters)
        est.coef[i] <- estimate.interpol.coef(i,y,e,est.coef)
      norm.coefs[e,y] <- norm.coefs[e,y] + sum(est.coef)
    }
  }

  cat("\nsolving\n")
  Im <- diag(m)
  #cat("final system ",Im - norm.coefs," ",Em,"\n")
  a[ind.na] <- solve(Im - norm.coefs,Em)
  attr(a,"reduce.df") <- m
  a
}

f.interpolate.bfaov <- function(anv,a) {
  # REQUIRE: anv$inters order compatible with is.subset order!
  ind.na <- which(is.na(a))
  m <- length(ind.na)
  Em <- coef(anv)[ind.na]
  inters <- anv$inters
  dims <- dim(a)
  dlist <- list()
  prod.dim.inters <- sapply(inters, function(d) 1/prod(dims[-d]))
  prod.dims <- 1/prod(dims)
  for (i in seq(along=dims))
    dlist <- c(dlist,list(1:dims[i]))
  na.coords <- do.call("expand.grid",dlist)[ind.na,]
  # conversion of data.frame to array for efficiency (big difference!)
  #na.m <- NULL
  #for (i in 1:nrow(na.coords))
  #  na.m <- cbind(na.m,as.numeric(na.coords[,i]))
  na.m <- as.matrix(na.coords)
  #print(na.m)

	len.inters <- length(inters)
  seq.inters <- seq(along=inters)
  is.sub.inters <- matrix(FALSE,length(inters),length(inters))
  # for efficient subset test:
  for (i in seq.inters)
    for (j in seq.inters)
      is.sub.inters[i,j] <- is.subset(inters[[i]],inters[[j]])
  diag(is.sub.inters) <- FALSE

  m.inters <- matrix.from.inters(inters)
  norm.coefs <- matrix(prod.dims,m,m)
  est.coefs <- rep(0,len.inters)

  norm.coefs <- matrix(.C("norm_coefs",
    as.integer(m),
    as.integer(t(m.inters)),
      as.integer(nrow(t(m.inters))),as.integer(ncol(t(m.inters))),
    as.integer(t(na.m)),
      as.integer(ncol(na.m)),as.integer(nrow(na.m)),
    as.double(prod.dim.inters),
      as.integer(length(prod.dim.inters)),
    as.double(prod.dims),
    as.logical(is.sub.inters),
      as.integer(nrow(is.sub.inters)),as.integer(ncol(is.sub.inters)),
    as.double(est.coefs),
      as.integer(length(est.coefs)),
    norm.coefs=as.double(norm.coefs),
      as.integer(nrow(norm.coefs)),
      as.integer(ncol(norm.coefs))
  )$norm.coefs,m,m)
  #cat("back\n")
  #print(norm.coefs)

  cat("\nsolving\n")
  Im <- diag(m)
  #cat("final system ",Im - norm.coefs," ",Em,"\n")
  a[ind.na] <- solve(Im - norm.coefs,Em)
  attr(a,"reduce.df") <- m
  a
}

      #rowv <- c(rowv,total.interpol.coef(y.ind,eq.ind,dims,inters))
total.interpol.coef <- function(y.ind,eq.ind,dims,inters) {
  # y.ind always contained in total mean
  c <- 1/prod(dims)
  # get coeffs within other model components
  for (i in seq(along=inters))
    c <- c + estimate.interpol.coeff(y.ind,eq.ind,inters[[i]],dims,inters)
  c
}

estimate.interpol.coeff <- function(y.ind,eq.ind,d,dims,inters) {
  # require length(d) > 0
  # if there is a match of indices in the relevant dimensions d:
  #cat("estimate.interpol on y ",y.ind," eq ",eq.ind," d ",d,"\n")
  if (all(y.ind[d] == eq.ind[d])) {
    c <- 1/prod(dims[-d])
    #cat("c for ",y.ind," in eq ",eq.ind," on d ",d," is ",c,"\n")
  }
  else
    c <- 0
  # recursively compute coefficients in contained models
  for (i in seq(along=inters)) {
    if (is.subset(inters[[i]],d) && !setequal(inters[[i]],d)) {
      c <- c - estimate.interpol.coeff(y.ind,eq.ind,inters[[i]],dims,inters)
    }
  }
  # empty list unfortunately not in inters list
  c <- c - 1/prod(dims)  # mu is always contained
  #cat("final c ",c," of ",y.ind," in eq ",eq.ind," on d ",d,"\n")
  c
}






# NEW STUFF ----------------------------------------------------








diff.g <- function(a) {
  gdim <- match("G",names(dim(a)))
  apply(a,gdim,mean)
}

diff.g.contr <- function(a,contr.var) {
  gdim <- match("G",names(dim(a)))
  cdim <- match(contr.var,names(dim(a)))
  m <- apply(a,c(gdim,cdim),mean)
  m[,1] - m[,2]
}

df.g <- function(a) {
  # degrees of freedom for t-distribution
  dims <- dim(a)
  prod(dims[-match(c("G"),names(dims))])
}

sd.g <- function(anv,formula=NULL) {
  # var of estimate should be
  # (1-1/dim[Gdim])*sigma.sq / prod(dim[-c(Gdim)]
  a <- anv$data
  dims <- dim(a)
  err <- anova(anv,formula,ptype="none")
  sqrt(as.numeric((1-1/dims[match("G",names(dims))])*err$MS/
        df.g(a) ))
}

hist.g <- function(anv,p.value=0.05,formula=NULL,a=NULL) {
  if (is.null(a) && !is.null(anv$data))
    a <- anv$data
  else
    stop("need array as parameter or in bfaov object")
  hist.g.v(anv,diff.g(a),p.value=p.value,formula=formula)
}

hist.g.v <- function(anv,g,p.value=0.05,formula=NULL) {
  q <- limit.g(anv,p.value=p.value,formula=formula)
  n <- dim(anv$data)["G"]
  #q <- abs(qnorm(p.value/n,0,1)*sd)
  hist(g,breaks=round(n/30),xlim=range(c(g,q,-q)),
      xlab="",ylab="",main="")
  lines(c(-q,-q),c(0,300),col="red",lwd=2)
  lines(c(q,q),c(0,300),col="red",lwd=2)
  q
}

limit.g <- function(anv,p.value=0.05,formula=NULL) {
  sd <- sd.g(anv,formula)
  df <- df.vg(anv$data)
  n <- dim(anv$data)["G"]
  #abs(qt(p.value/n,df)*sd)
  abs(qnorm(p.value/n,0,sd))
}

pval.g <- function(anv,over=T,formula=NULL,a=NULL) {
  if (is.null(a) && !is.null(anv$data))
    a <- anv$data
  else
    stop("need array as parameter or in bfaov object")
  sd <- sd.g(anv,formula)
  df <- df.g(anv$data)
  if (over) 
    #pt(diff.g(a)/sd*sqrt(df),df-1,lower.tail=F)
    pnorm(diff.g(a)/sd,lower.tail=F)
  else
    #pt(diff.g(a)/sd*sqrt(df),df-1)
    pnorm(diff.g(a)/sd)
}

pval.level <- function(anv,rml,over=T,a=NULL) {
  if (is.null(a) && !is.null(anv$data))
    a <- anv$data
  else
    stop("need array as parameter or in bfaov object")
  mv <- reml.mean.var("G",rml,anv$dims)
  sd <- sqrt(mv$sum)
  if (over) 
    pnorm(diff.g(a)/sd,lower.tail=F)
  else
    pnorm(diff.g(a)/sd)
}

pval.level.contr <- function(anv,rml,contr.var,over=T,a=NULL) {
  if (is.null(a) && !is.null(anv$data))
    a <- anv$data
  else
    stop("need array as parameter or in bfaov object")
  mv <- reml.mean.var(c("G",contr.var),rml,anv$dims)
  sd <- sqrt(2*mv$sum)
  if (over) 
    pnorm(diff.g.contr(a,contr.var)/sd,lower.tail=F)
  else
    pnorm(diff.g.contr(a,contr.var)/sd)
}

pval.g.boots <- function(b,a,over=T) {
  sdb <- if (!is.null(attr(b,"stats.only")))
      b[,2]
    else
      apply(b,1,sd)
  df <- df.g(a)
  if (over) 
    #pt(diff.g(a)/sdb*sqrt(df),df-1,lower.tail=F)
    pnorm(diff.g(a)/sdb,lower.tail=F)
  else
    #pt(diff.g(a)/sdb*sqrt(df),df-1)
    pnorm(diff.g(a)/sdb)

    #pnorm(0,diff.g(a),sdb)
    #pnorm(0,diff.g(a),sdb,lower.tail=F)
}
 

boots.g <- function(anv,n=50,formula=NULL,
    gloe=NULL,a=NULL,fast=T,stats.only=F,span=span) {
  if (is.null(a) && !is.null(anv$data))
    a <- anv$data
  else
    stop("need array as parameter or in bfaov object")
  gdim <- match("G",names(dim(a)))

  E <- coef(anv,formula)
  r <- a - E
  if (is.null(gloe)) {
    cat("loess fit to residuals, be patient ...\n")
    gloe <- get.loess(E,r,span=span)
  }

  pl <- predict.loess(gloe$loe,as.vector(E))
  loess.limit <- 0.05*median(pl)
  pl[pl <= loess.limit] <- loess.limit
  r.smooth <- sqrt(pl)
  r.scaled <- r/r.smooth
  if (!fast) {
    cat("bootstrapping",n,"rounds, be patient\n")
    cat("[consider fast option boots.g(...,fast=T)] ...\n")
    b <- NULL
    for (i in 1:n) {
      cat (i," ")
      mean.G <- apply(a,gdim,mean)
      b <- cbind( b, mean.G)
      a <- E + r.smooth*sample(r.scaled,repl=T)
    }
  } else {
    dr <- prod(dim(a)[-gdim])  # dim rest
    dg <- dim(a)[gdim]  # dim gene
    rest <- seq(along=dim(a))[-gdim]
    pa <- aperm(a,c(rest,gdim))
    pE <- aperm(E,c(rest,gdim))
    r.smooth <- array(r.smooth,dim(a))
    pr.smooth <- aperm(r.smooth,c(rest,gdim))
    r.scaled <- array(r.scaled,dim(a))
    pr.scaled <- aperm(r.scaled,c(rest,gdim))
    r <- array(r,dim(a))
    pr.orig <- aperm(r,c(rest,gdim))

    # for stats: b[,1] mean, b[,2] sd,
    # b[,3] count over, b[,4] count under
    bdim <- if (stats.only) 4 else n
    b <- matrix( .C("g_boots",
      as.integer(n),
      as.integer(dr),as.integer(dg),
      as.double(pE),
      #as.double(d.g),
      as.double(pr.smooth),as.double(pr.scaled),
      #as.double(rep(1,length(pr.orig))),as.double(pr.orig),
      as.double(rep(0,length(pr.orig))),
      #as.double(pr.orig),
      as.integer(stats.only),
      b=as.double(matrix(0,dg,bdim))
    )$b,dg,bdim)
  }
  if (stats.only) {
    attr(b,"stats.only") <- T
    b[,1] <- b[,1]/n  # mu 
    b[,2] <- sqrt(b[,2]/(n-1) - n*b[,1]^2/(n-1)) # sd
    # should follow a t-distr with df.g(a)-1 degs of freedom:
    b[,3] <- b[,3]/n  # p over
    b[,4] <- 1-b[,3]  # p under
  }
  cat ("\n")
  list(b=b,gloe=gloe)
}


get.diff.residuals <- function(a,E,E.base=NULL,gloe=NULL,span=NULL) {
  if (is.null(E.base))
    E.base <- E
  ng <- dim(a)["G"]
  r <- a - E
  if (is.null(gloe)) {
    cat("loess fit to residuals, be patient ...\n")
    gloe <- get.loess(E.base,r,span=span)
  }
  # basis for prediction, estimates one level below
  # possible alternative: gene level

  pl <- predict.loess(gloe$loe,as.vector(E.base))
  loess.limit <- 0.05*median(pl)
  pl[pl <= loess.limit] <- loess.limit
  r.smooth <- array(sqrt(pl),dim(a))
  #r.smooth <- array(1,dim(a))
  r.scaled <- r/r.smooth  
  list(r=r,
    r.smooth=r.smooth,
    r.scaled=r.scaled,
    gloe=gloe)
}

boots.level.g <- function(av,formulas,levels.seq,n=50,E.base=NULL,
    gloe.l=NULL,stats.only=F,span=span) {
  #formulas <- list(~G + S + A,
  #     ~G + S + A + G:S,~G + S + A + G:S + G:S:A)
  #levels.seq <- c("S","A")
  #n <- 100
  #old.gloe.l <- NULL
  #stats.only <- F

  old.gloe.l <- gloe.l
  gloe.l <- NULL

  a <- av$data

  levels.seq <- rev(levels.seq)
  level.inds <- sapply(c(levels.seq,"G"),grep,names(dim(a)))
  level.inds <- c(seq(along=dim(a))[-level.inds],level.inds)

  ng <- dim(a)["G"]
  nf <- length(formulas)
  E <- coef(av,formulas[[1]])
  r.l <- NULL
  r.m <- array(0,c(prod(dim(a))/ng,ng,nf))
  r.smooth.m <- r.m
  r.scaled.m <- r.m
  for (i in (1:nf) + 1) {
    if (i <= nf) {
      a.r <- coef(av,formulas[[i]])
      anova(av,formulas[[i]])
    } else {
      a.r <- a
    }
    if (is.null(old.gloe.l)) {
      res <- get.diff.residuals(a.r,E,E.base=E.base,span=span)
      gloe.l <- c(gloe.l,list(res$gloe))
    } else {
      res <- get.diff.residuals(a.r,E,E.base=E.base,
          old.gloe.l[[i-1]],span=span)
    }
    r.l <- c(r.l,list(res$r))
    r.m[,,i-1] <- matrix(aperm(res$r,level.inds),ncol=ng)
    r.smooth.m[,,i-1] <- matrix(aperm(res$r.smooth,level.inds),ncol=ng)
    r.scaled.m[,,i-1] <- matrix(aperm(res$r.scaled,level.inds),ncol=ng)
    E <- a.r
  }
  if (!is.null(old.gloe.l))
    gloe.l <- old.gloe.l

  # compare and test when rand48 is shut down
  # pA <- matrix(aperm(a,level.inds),ncol=ng)
  # all zero: ? 
  # pA - (apply(r.smooth.m*r.scaled.m,c(1,2),sum) + pG)
  # mA <- apply(pA,2,mean)

  gdim <- match("G",names(dim(a)))
  dr <- prod(dim(a)[-gdim])  # dim rest
  dg <- dim(a)[gdim]  # dim gene
  pa <- aperm(a,level.inds)
  E <- coef(av,formulas[[1]])
  G <- coef(av,~G)
  pE <- matrix(aperm(E,level.inds),ncol=ng)
  pG <- matrix(aperm(G,level.inds),ncol=ng)

  # spans: dim(a)[level.inds]: Q:2 A:2 S:3 G:5
  # cumprod and than take before S, before A, add 1
  # levels.seq: "A" "S"
  l.inds <- sapply(levels.seq,grep,names(dim(a)))
  spans <- c(cumprod(dim(a)[level.inds])[l.inds - 1],1)

  # for stats: b[,1] mean, b[,2] sd,
  # b[,3] count over, b[,4] count under
  bdim <- if (stats.only) 4 else n

  b <- matrix( .C("g_level_boots",
  #a.r <- matrix( .C("g_level_boots",
  #rc <- array( .C("g_level_boots",
      as.integer(n),
      as.integer(nf),
      as.integer(dr),
      as.integer(dg),
      eo=as.double(pE),
      #eo=as.double(array(0,dim(pE))),
      a.r=as.double(array(0,dim(pE))),
      as.double(r.smooth.m),
      as.double(r.scaled.m),
      #rc=as.double(array(0,dim(r.scaled.m))),
      as.double(pG[1,]),
      as.integer(spans),
      as.integer(stats.only),
      b=as.double(matrix(0,dg,bdim))
  )$b,dg,bdim)
  #)$a.r,ncol=ng)
  #)$rc,c(12,5,3))
  apply(a.r,2,mean)

  if (stats.only) {
  attr(b,"stats.only") <- T
    b[,1] <- b[,1]/n  # mu 
    b[,2] <- sqrt(b[,2]/(n-1) - n*b[,1]^2/(n-1)) # sd
    # should follow a t-distr with df.g(a)-1 degs of freedom:
    b[,3] <- b[,3]/n  # p over
    b[,4] <- 1-b[,3]  # p under
  }
  list(b=b,gloe.l=gloe.l)
}

boots.level.b.demo <- function(av,stats.only=T) {
  formulas <- list(~G + S + A,
      ~G + S + A + G:S,~G + S + A + G:S + G:S:A)
  levels.seq <- c("S","A")
  b <- boots.level.g(av,formulas,levels.seq,n=1000,
     stats.only=stats.only)

  if (!stats.only) 
    i <- 13; qqnorm(b[i,]); qqline(b[i,]); shapiro.test(b[i,]); ks.test(b[i,],"pnorm",mean=mean(b[i,]),sd=sd(b[i,]))

}


#/**
# \name{reml.bfaov}
# \alias{reml.bfaov}
# \alias{reml.kt}
# \alias{print.reml}
# \alias{optim.design}
# \title{REML estimates of variance components}
# \description{
#   REML estimates for balanced factorial designs
# }
# \usage{
#   reml.bfaov(anv,rand.facs,formula=NULL,zero=0.0001,anova.only=T)
#   reml.kt(ms,df,sf,A,zero=0.0001,anova=T)
#   print.reml(rml)
#   optim.design(anv,coords,w,limit=NULL,design=NULL)
# }
# \arguments{
#   \item{anv}{a bfaov object, see \code{\link{bfaov}}}
#   \item{rand.facs}{vector of random factors}
#   \item{formula}{a model formula, see \code{\link{bfaov}}}
#   \item{zero}{limit for assuming a variance component is zero}
#   \item{anova.only}{ANOVA estimates of variance components, can
#     result in negative components, but is fast}
#   \item{ms}{vector of mean squares}
#   \item{df}{vector of degrees of freedom}
#   \item{sf}{vector of coefficients of variance components in expected
#     mean square}
#   \item{A}{indicator matrix of variance components in expected
#     mean square}
#   \item{rml}{list returned by \code{reml.bfaov}}
#   \item{coords}{list of random main effects}
#   \item{w}{the cost vector for the random main effects}
#   \item{limit}{a limit on the costs of an experimental design}
#   \item{design}{vector describing hierarchical design: how
#     many of the first effect, how many of the second effect
#     nested in each of the first, etc.}
#   \item{anova}{if TRUE, negative MS estimates for variance components
#     are set to zero; otherwise, }
# }
# \value{
#   \code{reml.bfaov} returns a list of
#   \item{s.sq}{estimates of variance components}
#   \item{MS}{pseudo mean squares yielding the the variance components}
#   \item{error}{final error in the Kuhn-Tucker optimization}
#   \item{grad}{the gradient of the Kuhn-Tucker condition}
#   \item{coords}{list of symbols indicating the interaction effects}
# }
# \details{
#   If effects are assumed to be random the resulting mixed model
#   provides estimates of variance components instead of mean squares.
#   For balanced factorial designs the restricted maximum likelihood
#   estimates (REML) of the variance components can be calculated by
#   solving an optimization problem involving the ANOVA mean squares
#   (see W. A. Thompson, The problem of negative estimates of variance
#   components, Ann. of math. statistics 33:273-289, 1962).
# }
# \examples{
#   data(trcs)
#   RG <- rg.remove.quantile(trcs.RG,0.1)
#   RG <- rg.scale.normalize(RG)
#   a <- rg2log.array(RG,log.ratio=T)
#   av <- bfaov(~G + S/A + G:S + G:S:A,a)
#
#   # assume random effects S and A:
#   rml <- reml.bfaov(av,c("S","A"),~G + S/A + G:S + G:S:A)
#   print.reml(rml)
#   optim.design(av,c("S","A","Q"),c(50,10,1),limit=200)
#   
#   # W.A. Thompson example (Technometrics 5:441-449, 1963, p448)
#   
#   df <- c(3,1,3,8)
#   sf <- c(4,8,2,1)
#   ms <- c(5,8,9,4)
#   A <- rbind(c(1,0,1,1),c(0,1,1,1),c(0,0,1,1),c(0,0,0,1))
#   reml.kt(ms,df,sf,A,zero=0.01,anova=F)
# }
# \seealso{
#   \code{\link{bfaov}}
# }
# \keyword{models}
#*/

reml.bfaov <- function(anv,rand.facs,formula=NULL,zero=0.0001,
    anova.only=T) {
  require(MASS)
  comblist <- get.comblist(anv,formula)
  ac <- anova.from.comblist(anv,comblist)
  r.coords <- match(rand.facs,anv$coordnames)
  n <- ac$length
  inters <- ac$inters[1:(n-1)]
  rs <- sapply(inters,function(x,y)
      length(intersect(x,y)),r.coords)
  r.inds <- which(rs > 0)
  r.inters <- inters[r.inds]
  d <- length(r.inters)
  A <- matrix(0,d+1,d)
  for (i in 1:d) {
    for (j in i:d) {
      if (is.subset(r.inters[[i]],r.inters[[j]]))
        A[i,j] <- 1
    }
  }
  A <- cbind(A,rep(1,d+1))
  r.MS <- c(ac$MS[r.inds],ac$MS[n])
  r.df <- c(ac$df[r.inds],ac$df[n])
  #s.sq <- solve(A,r.MS)
  r.sf <- sapply(r.inters,function(inter) prod(anv$dims[-inter]))
  r.sf <- c(r.sf,1)
  r.names <- sapply(r.inters,
      function(inter) anv$coordnames[inter] )
  r.names <- c(r.names,list("res"))
  if (anova.only) {
    c(var.comp.anova(r.MS,r.df,r.sf,A),list(coords=r.names,inds=r.inds))
  } else {
    c(reml.kt(r.MS,r.df,r.sf,A,zero=zero),list(coords=r.names,inds=r.inds))
  }
}

reml.kt <- function(ms,df,sf,A,zero=0.0001,anova=T) {
  A.i <- ginv(A)
  B <- NULL
  for (i in 1:nrow(A))
    B <- rbind(B,A[i,] * sf)
  #B <- rbind(c(1,0,0,0),c(1,4,0,0),c(1,0,20,0),c(1,4,20,40))
  B.i <- ginv(B)

  var.kuhn.tucker.sub <- function(x,inds,ms.max=1,ms.abs=1,ms.c=1) {
    c1 <- A.i %*% x
    c2 <- t(A) %*% (df*(ms-x)/x^2) 
    ( abs(sum(df) - sum(df*ms/x)) * ms.c +
      ( sum(pmax(0,-c1[-inds])) + sum(pmax(0,c2[inds])) ) * ms.max +
      ( sum(abs(c1[inds])) + sum(abs(c2[-inds])) ) * ms.abs
    )
  }

  var.kuhn.tucker <- function(x,ms.max=1,ms.c=1) {
    c1 <- A.i %*% x
    c2 <- t(A) %*% (df*(ms-x)/x^2) 
    ( abs(sum(df) - sum(df*ms/x))*ms.c +
      (sum(pmax(0,-c1)) + sum(pmax(0,c2)))*ms.max )
  }

  p <- NULL
  p$par <- ms
  for (i in 1:20)
    p <- optim(p$par,var.kuhn.tucker,ms.max=100)
  cat(".")
  for (i in 1:20)
    p <- optim(p$par,var.kuhn.tucker,ms.max=10)
  cat(".")
  for (i in 1:20)
    p <- optim(p$par,var.kuhn.tucker,ms.max=1)
  cat(".")

  s.sq <- B.i %*% p$par
  s.sq.ms <- B.i %*% ms
  #inds <- which(s.sq < zero.frac*max(s.sq))
  inds <- which(s.sq < zero)
  inds.ms <- which(s.sq.ms < zero)
  if (!setequal(inds,inds.ms)) {
    if (anova) {
      warning("negative MS var comps differ from < zero var comps\ntaking MS")
      inds <- inds.ms
    } else {
      warning("negative MS var comps differ from < zero var comps\ntaking < zero")
    }
  }
  #cat("inds ",inds,"\n")
  #inds <- c(1,3)

  if (length(inds) > 0) {
    for (i in 1:20)
      p <- optim(p$par,var.kuhn.tucker.sub,
          control=list(abstol=1e-15,reltol=1e-15),
          inds=inds,ms.max=100 );
    cat(".")
    for (i in 1:20)
      p <- optim(p$par,var.kuhn.tucker.sub,
          control=list(abstol=1e-15,reltol=1e-15),
          inds=inds,ms.max=10 )
    cat(".")
    for (i in 1:20)
      p <- optim(p$par,var.kuhn.tucker.sub,
          control=list(abstol=1e-15,reltol=1e-15),
          inds=inds,ms.max=1 )
    cat(".\n")
  }
  list(s.sq=B.i %*% p$par,MS=p$par,s.sq.ms=s.sq.ms,
      error=p$value,grad=df*(ms-p$par)/p$par^2)
}

var.comp.anova <- function(ms,df,sf,A) {
  B <- NULL
  for (i in 1:nrow(A))
    B <- rbind(B,A[i,] * sf)
  B.i <- ginv(B)
  s.sq.ms <- B.i %*% ms
  list(s.sq=B.i %*% ms,MS=ms,
      error=0,grad=0)
}

print.reml <- function(rml) {
  for (i in seq(along=rml$s.sq)) {
    if (length(rml$coords[[i]]) > 1) 
      for (j in 1:(length(rml$coords[[i]])-1)) {
        cat((rml$coords[[i]])[j],",",sep="")
      }
    cat(rml$coords[[i]][length(rml$coords[[i]])],"\t",
        mformat(rml$s.sq[i]),"\n")
  }
}

reml.mean.var <- function(mean.coords,rml,dims) {
  sum <- 0
  str <- NULL
  s.part <- NULL
  for (i in seq(along=rml$s.sq)) {
    if (rml$coords[[i]] == "res")
      rml$coords[[i]] <- names(dims)
    coords <- setdiff(rml$coords[[i]],mean.coords)
    s.part <- c(s.part,rml$s.sq[i]/prod(dims[coords]))
    sum <- sum + rml$s.sq[i]/prod(dims[coords])
    str <- c(str,
        paste(mformat(rml$s.sq[i]),
        "/(",paste(coords,collapse="*"),")",sep="" ) )
  }
  list(sum=sum,str=paste(str,collapse=" + "),s.part=s.part)
}


optim.mean.var <- function(str,coords,w,limit=NULL,design=NULL) {
  costfunc <- function(v) {
    sum <- 0
    for (i in seq(along=v))
      sum <- sum + prod(v[1:i])*w[i]
    sum
  }
  if (is.null(limit) && !is.null(design))
    limit <- costfunc(design)
  m <- matrix(1:floor(limit/w[1]),ncol=1)
  if (length(w) > 1) {
    for (i in 2:length(w)) {
      mm <- NULL
      for (j in 1:nrow(m)) {
        c <- prod(m[j,1:(i-1)])
        rest.limit <- floor((limit-costfunc(m[j,1:(i-1)]))/(c*w[i]))
        if (rest.limit >= 1) {
          if (i < length(w))
            for (k in 1:rest.limit)
              mm <- rbind(mm,c(m[j,],k))
          else
              mm <- rbind(mm,c(m[j,],rest.limit))
        }
      }
      m <- mm
    }
  }   
  f <- eval(parse(
      text=paste("function(",paste(coords,collapse=","),")",str)) )
  all.costs <- apply(m,1,costfunc)
  inds <- which(all.costs <= limit)
  cases <- m[inds,]
  costs <- all.costs[inds]
  s.sq <- apply(cases,1,function(v) do.call("f",as.list(v)))
  best <- which(s.sq == min(s.sq))
  names(s.sq) <- NULL
  list(case=cases[best,],cost=costs[best],sq=s.sq[best],limit=limit)
}

optim.design <- function(anv,coords,w,limit=NULL,design=NULL) {
  mv <- reml.mean.var("G",rml,anv$dims)
  optim.mean.var(mv$str,coords,w,limit=limit,design=design)
}



# -------------------------------------


# Add documentation for boots later:

#
# \name{boots.vg}
# \alias{boots.vg}
# \alias{boots.g}
# \alias{boots.level.g}
# \title{Significance of variety-gene effects by bootstrapping of residuals}
# \description{
#   Bootstrap residuals to estimate the significance of over- and
#   under-expression from an ANOVA model \code{\link{bfaov}}.
# }
# \usage{
#   boots.vg(anv,n=50,formula=NULL,
#     gloe=NULL,a=NULL,fast=T,stats.only=F)
#   boots.g(anv,n=50,formula=NULL,
#     gloe=NULL,a=NULL,fast=T,stats.only=F)
#   boots.level.g(av,formulas,levels.seq,n=50,E.base=NULL,
#     gloe.l=NULL,stats.only=T)
# \arguments{
#   \item{anv}{a bfaov object, containing a V and G effect for
#     \code{boots.vg}, and a G effect only (log ratios) for \code{boots.g}
#     and \code{boots.level.g}, see \code{\link{bfaov}}}
#   \item{n}{number of bootstrap replicates}
#   \item{formula}{a model formula, see \code{\link{bfaov}}}
#   \item{formulas}{a list of a model formulas, describing
#     the residuals for hierarchical bootstrap,
#     see \code{\link{bfaov}}}
#   \item{levels.seq}{a list of main effects, the levels for hierarchical
#     bootstrap}
#   \item{gloe}{a loess object from a smooth regression to the
#     residuals (over estimates), for example, from an earlier call to
#     \code{boots.vg}, see also
#     \code{\link{loess}} }
#   \item{gloe.l}{a list of loess objects as in \code{gloe} for hierarchical
#     bootstrap}
#   \item{a}{a balanced factorial data array, see
#     \code{\link{rg2log.array}} }
#   \item{fast}{if set to T, the compiled bootstrap is called which is
#     \emph{much} faster then the R version, but might not be
#     available on every platform }
#   \item{span}{span for loess smoother}
#   \item{stats.only}{if TRUE, only mean and standard deviation of
#     VG or G effect is collected for each bootstrap replicate}
# }
# \value{
#   \item{b}{a matrix of bootstrap results for simulating VG, G,
#     or level effects.
#     \code{b[,1]} is the vector of the original effect for every
#     gene, \code{b[,i]} is the vector of the \eqn{i}{i}-th simulation
#   } 
#   \item{gloe}{a loess object from a smooth regression to the
#     residuals (over estimates), used to scale the residuals for the
#     bootstrapping }
# }
# \details{
#   Usually the residuals for microarrays are not normally distributed
#   and can show heteroscedasticity (dependency of residuals on fitted
#   values). The estimates of significance derived from the residual
#   error (see \code{\link{sd.vg}}) might not be reliable enough
#   (although they often underestimate significance and are
#   conservative). Bootstrapping redistributes the residuals over the
#   fitted values and recalculates estimates for VG or G effects in
#   each round. This estimates provide a saver basis for calculating
#   p-values of significant over- or under-expression. \code{boots.vg}
#   bootstraps the VG effect, \code{boots.g} the gene effect G for log
#   ratios, and \code{boots.level.g} log ratio gene effects G in a
#   hierarchical fashion.
# }
# \examples{
#   data(trcs)
#   bvg <- boots.vg(trcs.av,n=50,stats.only=F)
# 
#   # are bootstrap VG effects for gene 30 normal?
#   qqnorm(bvg$b[30,])
#   qqline(bvg$b[30,])
#
#   RG <- rg.remove.quantile(trcs.RG,0.1)
#   a <- rg2log.array(RG,log.ratio=T)
#   av <- bfaov(~G + S/A + G:S + G:S:A,a)
#
#   # 3 levels of bootstrap: samples S, arrays A, residuals:
#   formulas <- list(~G + S + A,
#       ~G + S + A + G:S,~G + S + A + G:S + G:S:A)
#   levels.seq <- c("S","A")
#   blg <- boots.level.g(av,formulas,levels.seq,n=200,
#       stats.only=F,span=0.4)
#   x <- blg$b[30,]; qqnorm(x); qqline(x); ks.test(x,"pnorm",mean(x),sd(x))
#
#   # plot loess fit for residuals:
#   opar <- par(mfrow=c(3,1))
#   for (i in 1:3) plot.loess(blg$gloe.l[[i]])
#   par(opar)
# }
# \seealso{
#   \code{\link{over.under}, \code{\link{over.under.level}}
# }
# \keyword{models}
#

