# The original version of YASMA is due to L. Wernisch
# at http://people.cryst.bbk.ac.uk/~wernisch/yasma.html
# This version has three patches which were done 
# by P. Sykacek in 2005 (January -May):
# 1) change to predict for loess objects since the original
#    version of predict.loess was incompatible with R >= 1.9
# 2) change the original reference to various stats libraries to
#    stats (The original version gave warnings, since core R changed
#    it's library layout.)
# 3) change drand48 in yasma.c to my_rand, which is based tkes into account
#    that drand48() is not strict ANSI and not awailable with the tools
#    suggested to port R libs to Win32. This one either uses drand48() or
#    rand() if the former is not available. The original c-code did work under
#    win32 but only with a special cygwin1.dll installed, which made a
#    generic Win32 distribution impossible.
#    comments of any kind: ps408@cam.ac.uk (Peter Sykacek)
# P.S., May 2005
library(stats)

#/**
# @name
# \title{
#   RG structure from microarray data table file
# }
# \description{
#   Parses files containing microarray data and stores them
#   in the RG format (see \code{\link{init.data}} from package sma).
#   Various patterns describing table headings and lines to skip 
#   must be provided.
# }
# \usage{
#   read.rg(directory="",
#     R.filenames,R.prefix="",R.suffix="",
#     x.col=NULL,y.col=NULL,
#     G.filenames,G.prefix="",G.suffix="",
#     gene.col,R.col,Rb.col,G.col,Gb.col,
#     do.prepare=T,split.char="\t",
#     start.phrase=NULL,end.phrase=NULL,
#     experiments=NULL)
# }
# @synopsis
# \arguments{
#   \item{directory}{data directory}
#   \item{R.filenames}{a list of filenames containing data of "red" R
#     channel without prefix or suffix}
#   \item{R.prefix, R.suffix}{prefix and suffix for R data files}
#   \item{gene.col}{column of gene names}
#   \item{R.col, Rb.col}{signal and background column name}
#   \item{x.col, y.col}{spot coordinates}
#   \item{G.filenames, G.prefix, G.suffix, G.col, Gb.col}{%
#      the same as above but for the "green" G channel}
#   \item{do.prepare}{if TRUE space is replaced by "." in order to
#     use \R's read.table() function}
#   \item{split.char}{the character used as table separator}
#   \item{start.phrase}{a phrase in the table header line, if set all previous
#     lines are skipped in the input files}
#   \item{end.phrase}{a phrase in the line after the last table line,
#     if set the rest of the input file is skipped}
#   \item{experiments}{a list of named vectors describing the content
#   of the experiments; eg, list(P=1:3,A=1:2) means 3 preparations
#   with 2 arrays each (total 9): 1-2 with mRNA preparation
#   (P) I, 3-4 with preparation II, and 5-6 with prepartion III}
# }
# \details{
#   Unfortunately, there exists no standard for storing microarray
#   scanner data yet. This function helps to parse data from a variety
#   of data formats. One common problem is that R's read.table()
#   function is confused by spaces (eg in gene names), they are
#   replaced by ".". Text before the proper table headers and after
#   the tables can be skipped as well.
# }
# \value{
#   an RG structure, see \code{\link{init.data}} from package sma
# }
# \examples{
#   \dontrun{
#    my.RG <- read.rg("/home/rtp/microarrays/data/mtb/",
#      gene.col="Gene ID",
#      R.filenames=c("01cy5","03cy5","05cy5",
#        "02cy5","04cy5","06cy5" ),
#      R.prefix="MT-",R.suffix=".dat",
#      G.filenames=c("01cy3","03cy3","05cy3",
#        "02cy3","04cy3","06cy3" ),
#      G.prefix="MT-",G.suffix=".dat",
#      R.col="Signal Median",Rb.col="Background Median",
#      G.col="Signal Median",Gb.col="Background Median",
#      do.prepare=T,start.phrase="Gene ID",end.phrase="End Raw Data",
#      experiments=list(P=1:3,A=1:2)
#    )
#   }
# }
# \seealso{
#   \code{\link{init.data}}
# }
# \keyword{file}
#*/

# test of my random stuff
my.rand <- function()
  .C("rand_interface",
     ret=as.double(0))$ret

read.rg <- function(directory="",
  R.filenames,R.prefix="",R.suffix="",
  G.filenames,G.prefix="",G.suffix="",
  gene.col,R.col,Rb.col,G.col,Gb.col,
  name.col=NULL,
  Rs.col=NULL,Rbs.col=NULL,Rsn.col=NULL,Rbsn.col=NULL,
  Gs.col=NULL,Gbs.col=NULL,Gsn.col=NULL,Gbsn.col=NULL,
  x.col=NULL,y.col=NULL,
  do.prepare=T,split.char="\t",
  start.phrase=NULL,end.phrase=NULL,
  experiments=NULL ) {

  r.col.v <- c(gene.col,R.col,Rb.col)
  if (!is.null(name.col))
    r.col.v <- c(r.col.v,name.col)
  if (!is.null(x.col) && !is.null(y.col)) {
    r.col.v <- c(r.col.v,x.col,y.col)
  }
  if (!is.null(Rs.col) && !is.null(Rbs.col)
    && !is.null(Rsn.col) && !is.null(Rbsn.col) ) {
    r.col.v <- c(r.col.v,Rs.col,Rbs.col,Rsn.col,Rbsn.col)
  }
  g.col.v <- c(gene.col,G.col,Gb.col)
  if (!is.null(Gs.col) && !is.null(Gbs.col) &&
      !is.null(Gsn.col) && !is.null(Gbsn.col) ) {
    g.col.v <- c(g.col.v,Gs.col,Gbs.col,Gsn.col,Gbsn.col)
  }
  paste.core <- function(core,prefix,suffix)
    paste(c(directory,prefix,core,suffix),
        sep="",collapse="")
  r.files <- sapply(R.filenames,paste.core,prefix=R.prefix,suffix=R.suffix)
  g.files <- sapply(G.filenames,paste.core,prefix=G.prefix,suffix=G.suffix)
  r.cols <- get.columns(r.files,r.col.v,
      do.prepare=do.prepare,start.phrase=start.phrase,end.phrase=end.phrase)
  g.cols <- get.columns(g.files,g.col.v,
      do.prepare=do.prepare,start.phrase=start.phrase,end.phrase=end.phrase)
  R <- NULL; Rb <- NULL;
  Rs <- NULL; Rbs <- NULL;
  Rsn <- NULL; Rbsn <- NULL;
  G <- NULL; Gb <- NULL;
  Gs <- NULL; Gbs <- NULL;
  Gsn <- NULL; Gbsn <- NULL;
  x <- NULL; y <- NULL;
  genes <- as.vector(r.cols[[1]][,match(gene.col,r.col.v)]);
  if (!is.null(name.col)) {
    names <- as.vector(r.cols[[1]][,match(name.col,r.col.v)]);
    genes <- paste(genes,names)
  }
  for (i in seq(along=r.cols)) {  
    R <- cbind(R,as.numeric(r.cols[[i]][,match(R.col,r.col.v)]))
    Rb <- cbind(Rb,as.numeric(r.cols[[i]][,match(Rb.col,r.col.v)]))
    if (!is.null(Rs.col)) {
      Rs <- cbind(Rs,as.numeric(r.cols[[i]][,match(Rs.col,r.col.v)]))
      Rsn <- cbind(Rsn,as.numeric(r.cols[[i]][,match(Rsn.col,r.col.v)]))
      Rbs <- cbind(Rbs,as.numeric(r.cols[[i]][,match(Rbs.col,r.col.v)]))
      Rbsn <- cbind(Rbsn,as.numeric(r.cols[[i]][,match(Rbsn.col,r.col.v)]))
      R <- cbind(Rbs,as.numeric(r.cols[[i]][,match(Rsn.col,r.col.v)]))
    }
    if (!is.null(x.col) && !is.null(y.col)) {
      x <- cbind(x,as.numeric(r.cols[[i]][,match(x.col,r.col.v)]))
      y <- cbind(y,as.numeric(r.cols[[i]][,match(y.col,r.col.v)]))
    }
    G <- cbind(G,as.numeric(g.cols[[i]][,match(G.col,g.col.v)]))
    Gb <- cbind(Gb,as.numeric(g.cols[[i]][,match(Gb.col,g.col.v)]))
    if (!is.null(Gs.col)) {
      Gs <- cbind(Gs,as.numeric(g.cols[[i]][,match(Gs.col,g.col.v)]))
      Gsn <- cbind(Gsn,as.numeric(g.cols[[i]][,match(Gsn.col,g.col.v)]))
      Gbs <- cbind(Gbs,as.numeric(g.cols[[i]][,match(Gbs.col,g.col.v)]))
      Gbsn <- cbind(Gbsn,as.numeric(g.cols[[i]][,match(Gbsn.col,g.col.v)]))
    }
  }
  list(genes=genes,experiments=experiments,
    R=R,Rb=Rb,G=G,Gb=Gb,x=x,y=y,Rs=Rs,Rbs=Rbs,Gs=Gs,Gbs=Gbs)
}

blank.subs <- "."

prepare.file <- function(filename,first.phrase=NULL,last.phrase=NULL) {
  fd <- file(filename)
  txt <- readLines(fd,-1)
  ## potential connection waste, use
  close(fd)
  start <- ifelse (!is.null(first.phrase),
      grep(first.phrase,txt)[1], 1 )
  end <- ifelse (!is.null(last.phrase),
      grep(last.phrase,txt)[1] - 1, length(txt) )
  txt <- txt[start:end]
  txt <- gsub(" ",blank.subs,txt)
  #tmp.file <- paste(filename,".prep",sep="")
  tmp.file <- "tmp.dat"
  fd.out <- file(tmp.file,open="w")
  writeLines(txt,fd.out)
  close(fd.out)
  tmp.file
}

get.columns <- function(filenames,column.names,
    do.prepare=F,start.phrase=NULL,end.phrase=NULL,split.char="\t") {
  column.names <- gsub(" ",blank.subs,column.names)
  lst <- list()
  for (i in seq(along=filenames)) {
    if (do.prepare)
      t <- read.table(prepare.file(filenames[i],start.phrase,end.phrase),
          header=T )
    else
      t <- read.table(filenames[i],header=T)
    lst <- c(lst,list(t[,column.names]))
  }
  lst
}

#/**
# @name
# \title{
#   Log2 matrix from RG structure
# }
# \description{
#   Returns a matrix of (background corrected) log ratios of \eqn{R/G}{R/G}
# }
# @usage
# @synopsis
# \arguments{
#   \item{RG}{RG microarray data structure, see \code{\link{read.rg}}}
# }
# \value{
#   Matrix of \eqn{log_2((R-Rb)/(G-Gb))}{log_2((R-Rb)/(G-Gb))}
#   (background corrected log2 ratio) values with genes as rows and
#   experiments as columns.
# }
# \examples{
#   data(trcs)
#   al <- rg.log.matrix(trcs.RG)
#   cor(al)
# }
# \keyword{manip}
#*/

rg.log.matrix <- function(RG) {
  r <- RG$R-RG$Rb
  g <- RG$G-RG$Gb
  r[r <= 0] <- 1
  g[g <= 0] <- 1
  log(r,2)-log(g,2)
}

#/**
# \name{rg.remove}
# \alias{rg.remove}
# \alias{rg.keep}
# \alias{rg.remove.containing}
# \alias{rg.keep.containing}
# \alias{rg.project}
# \alias{rg.write}
# \title{Removal of genes and experiments from RG structure}
# \description{
#   Removes genes from RG structure, either by index or by a pattern
#   describing gene names.
# }
# \usage{
#   rg.keep(RG,ind)
#   rg.remove(RG,ind)
#   rg.keep.containing(RG,patterns)
#   rg.remove.containing(RG,patterns)
#   rg.project(RG,dim=NULL,experiments=NULL)
#   rg.write(RG,prefix="rg")
# }
# \arguments{
#   \item{RG}{RG microarray data structure}
#   \item{ind}{index vector of genes to keep or remove}
#   \item{patterns}{vector of regular expression describing gene names}
#   \item{dim}{indices of experiments to keep}
#   \item{experiments}{the new experiment structure after projection}
#   \item{prefix}{prefix of filenames to write the RG structure}
# }
# \value{
#   RG structure
# }
# \details{
#   \code{rg.keep} and \code{rg.remove} keep or remove genes indexed
#   by \code{ind} from an RG structure. \code{rg.keep.containing} and
#   \code{rg.keep.remove} keep or remove genes containing a regular
#   expression pattern, see \code{\link{grep}}. \code{rg.project} confines
#   the RG structure to a subset of experiments given by \code{dim}.
#   \code{rg.write} writes an RG structure to a file.
# }
# \examples{
#   data(trcs)
#   RG <- rg.keep.containing(trcs.RG,"Rv")
#   RG.sample1 <- rg.project(RG,1:4,experiments=list(A=1:2,Q=1:2))
#   rg.write(RG.sample1,"sample1.dat")
# }
# \seealso{
#   \code{\link{read.rg}}, \code{\link{init.data}}
# }
# \keyword{manip}
#*/

rg.keep <- function(RG,ind) {
  if (length(ind) > 0) {
    RG$genes <- RG$genes[ind]
    RG$R <- RG$R[ind,,drop=F]
    RG$Rb <- RG$Rb[ind,,drop=F]
    RG$G <- RG$G[ind,,drop=F]
    RG$Gb <- RG$Gb[ind,,drop=F]
    if (length(RG$x) > 0 && length(RG$y) > 0) {
      RG$x <- RG$x[ind,,drop=F]
      RG$y <- RG$y[ind,,drop=F]
    }
  }
  RG
}

rg.remove <- function(RG,ind) {
  if (length(ind) > 0) {
    RG$genes <- RG$genes[-ind]
    RG$R <- RG$R[-ind,,drop=F]
    RG$Rb <- RG$Rb[-ind,,drop=F]
    RG$G <- RG$G[-ind,,drop=F]
    RG$Gb <- RG$Gb[-ind,,drop=F]
    if (length(RG$x) > 0 && length(RG$y) > 0) {
      RG$x <- RG$x[-ind,,drop=F]
      RG$y <- RG$y[-ind,,drop=F]
    }
  }
  RG
}

rg.index.containing <- function(RG,patterns) {
  as.vector(unlist(sapply(patterns,grep,RG$genes)))
  #ind <- NULL
  #for (i in seq(along=patterns))
  #  ind <- union(ind,grep(patterns[i],RG$genes))
  #ind
}

rg.keep.containing <- function(RG,patterns) {
  rg.keep(RG,rg.index.containing(RG,patterns))
}

rg.remove.containing <- function(RG,patterns) {
  rg.remove(RG,rg.index.containing(RG,patterns))
}

rg.project <- function(RG,dim=NULL,experiments=NULL) {
  if (length(dim) > 0) {
    RG$R <- RG$R[,dim,drop=F]
    RG$Rb <- RG$Rb[,dim,drop=F]
    RG$G <- RG$G[,dim,drop=F]
    RG$Gb <- RG$Gb[,dim,drop=F]
    if (length(RG$x) > 0 && length(RG$y) > 0) {
      RG$x <- RG$x[,dim,drop=F]
      RG$y <- RG$y[,dim,drop=F]
    }
    if (!is.null(experiments))
      RG$experiments <- experiments
  }
  RG
}

rg.merge <- function(RG1,RG2,experiments=NULL) {
  RG <- NULL
  RG$genes <- RG1$genes
  RG$R <- cbind(RG1$R,RG2$R)
  RG$Rb <- cbind(RG1$Rb,RG2$Rb)
  RG$G <- cbind(RG1$G,RG2$G)
  RG$Gb <- cbind(RG1$Gb,RG2$Gb)
  if (length(RG1$x) > 0 && length(RG2$y) > 0) {
    RG$x <- cbind(RG1$x,RG2$y)
    RG$y <- cbind(RG1$x,RG2$y)
  }
  if (!is.null(experiments))
    RG$experiments <- experiments
  else
    RG$experiments <- c(list(E=1:2),wc.RG$exp)
  RG
}

rg.write <- function(RG,prefix) {
  for (i in 1:ncol(RG$R)) {
    dfr <- data.frame(genes=RG$genes,
      signal=RG$R[,i],
      bg=RG$Rb[,i] )
    write.table(dfr,file=paste(prefix,"-r-",i,sep=""),sep="\t",
      quote=F,row.names=F)
    dfr <- data.frame(genes=RG$genes,signal=RG$G[,i],
      bg=RG$Gb[,i] )
    write.table(dfr,file=paste(prefix,"-g-",i,sep=""),sep="\t",
      quote=F,row.names=F)
  }
}


#/**
# \name{rg.zero.bg}
# \alias{rg.zero.bg}
# \alias{rg.sub.bg}
# \alias{rg.add2bg}
# \title{Background manipulation}
# \description{
#   Subtracts background from signal, adds to background, or sets
#   background to zero
# }
# \usage{
#   rg.zero.bg(RG)
#   rg.sub.bg(RG,min.val=1)
#   rg.add2bg(RG,r,g=NULL)
# }
# \arguments{
#   \item{RG}{RG microarray data structure, see \code{\link{read.rg}}}
#   \item{min.val}{negative values (background higher than signal) 
#       are reset to min.val}
#   \item{r,g}{values to add to background, if only \code{r} is given
#       \code{g} is set to \code{r}}
# }
# \value{
#   RG structure
# }
# \details{
#   \code{rg.sub.bg} subtracts background from signal and sets
#   background to 0, \code{rg.zero.bg} sets background to zero,
#   \code{rg.add2bg} adds constants to background
# }
# \examples{
#   data(trcs)
#   # reduce background values
#   RG <- rg.add2bg(trcs.RG,r=-500,g=-500)
#   RG <- rg.sub.bg(RG)
# }
# \seealso{
#   \code{\link{read.rg}}
# }
# \keyword{manip}
#*/

rg.sub.bg <- function(RG,min.val=1) {
  RG$R <- RG$R - RG$Rb
  RG$Rb <- array(0,dim(RG$Rb))
  RG$G <- RG$G - RG$Gb
  RG$Gb <- array(0,dim(RG$Gb))
  if (min.val >= 0) {
    RG$R[RG$R <= 0] <- min.val
    RG$G[RG$G <= 0] <- min.val
  }
  RG
}

rg.zero.bg <- function(RG) {
  RG$Rb <- array(0,dim(RG$Rb))
  RG$Gb <- array(0,dim(RG$Gb))
  RG
}

rg.add2bg <- function(RG,r,g=NULL) {
  if (is.null(g)) g <- r
  RG$Rb <- RG$Rb + r
  RG$Gb <- RG$Gb + g
  RG
}

#/**
# @name
# \title{Transforms multiple spots into extra experiments}
# \description{
#   Transforms RG structure by adding extra experiments for 
#   multiple genes on one array
# }
# @usage
# @synopsis
# \arguments{
#   \item{RG}{RG microarray data structure, see \code{\link{read.rg}}}
# }
# \details{
#   If \code{RG$genes} contains multiple entries for genes, extra
#   experiments for these multiple entries are added to \code{RG}. For
#   example, if there are 2 genes per array and 6 experiments, after
#   addition there are 12 experiments with 1,2 containing the two
#   genes from the first experiment, 3,4 from the second, etc
# }
# \value{
#   RG structure with genes \code{ind} with experiments added
# }
# \examples{
#   data(trcs)
#   # generate pseudo second spot on each array:
#   RG <- trcs.RG
#   dim(RG$R)
#   for (i in seq(along=names(RG))[-grep("experiments",names(RG))]) {
#     RG[[i]] <- rbind(RG[[i]],RG[[i]])
#   }
#   dim(RG$R)
#   RG <- rg.duplicate.genes(RG)
#   dim(RG$R)
# }
# \seealso{
#   \code{\link{read.rg}}, \code{\link{init.data}}
# }
# \keyword{manip}
#*/

rg.duplicate.genes <- function(RG) {
  dup <- split.unique(RG$genes)
  R <- NULL; Rb <- NULL;
  G <- NULL; Gb <- NULL;
  x <- NULL; y <- NULL;
  for (i in 1:ncol(RG$R)) {
    for (k in 1:ncol(dup)) {
      R <- cbind(R,RG$R[dup[,k],i])
      Rb <- cbind(Rb,RG$Rb[dup[,k],i])
      G <- cbind(G,RG$G[dup[,k],i])
      Gb <- cbind(Gb,RG$Gb[dup[,k],i])
      if (length(RG$x) > 0) {
        x <- cbind(x,RG$x[dup[,k],i])
        y <- cbind(y,RG$y[dup[,k],i])
      }
    }
  }
  experiments <- c(RG$experiments,list(D=1:ncol(dup)))

  list(genes=RG$genes[dup[,1]],
      R=R,Rb=Rb,G=G,Gb=Gb,x=x,y=y,
      experiments=experiments )
}

split.unique <- function(genes) {
  ind1 <- NULL; ind2 <- NULL
  n <- length(genes)
  for (i in 1:(n-1)) {
    j <- match(genes[i],genes[(i+1):n])
    if (!is.na(j)) {
      ind1 <- c(ind1,i)
      ind2 <- c(ind2,i+j)
    }
  }
  cbind(ind1,ind2)
}

#/**
# @name
# \title{Removal of genes with low expression levels}
# \description{
#     Removes genes which fall in the lower quantile in a specified
#     number of experiments
# }
# \usage{
#   rg.remove.quantile(RG,p.low=0.2,level=1,p.high=1,
#     intensities=T,set.na=F)
# }
# @synopsis
# \arguments{
#   \item{RG}{RG microarray data structure, see \code{\link{read.rg}}}
#   \item{p.low}{the lower quantile to remove from all channels}
#   \item{level}{minimum number of experiments for which data
#     point must lie in
#     the given quantile to be removed}
#   \item{p.high}{the upper quantile to remove from all channels}
#   \item{intensities}{if TRUE, spot intensities (R*G), otherwise
#     separate R, G channel values are used}
#   \item{set.na}{if TRUE and if \code{level} > 1 genes with low
#   values in less than \code{level} experiments are kept but such
#   values are marked with NA for later treatment in
#   \code{\link{interpolate.bfaov}}}
# }
# \details{
#   Genes with their value in the lower \code{p.low} quantile of
#   \code{level} or more R or G channels (or total intensity R*G) of
#   the experiments are removed. If \code{set.an=T} genes in quantiles
#   of fewer than \code{level} channels are set to \code{NA} in these
#   channels. They can be interpolated by
#   \code{\link{interpolate.bfaov}}. For example, if \code{level} is 2
#   and \code{p.low} is 0.1, all genes with values in the lower
#   10\% in two or more experiments are removed completely; if
#   \code{set.na=T}, all genes with a value in the lower 10\% of only
#   one experiment have such value replaced by \code{NA};
# }
# \value{
#   an RG structure, see \code{\link{init.data}} from package sma
# }
# \examples{
#   data(trcs)
#   RG <- rg.remove.quantile(trcs.RG,0.1,level=2,set.na=T)
#   # number of missing vaues:
#   sum(is.na(RG$R))
#   # the same as
#   sum(is.na(RG$G))
# }
# \seealso{
#   \code{\link{read.rg}}, \code{\link{rg.failure.histogram}},
#   \code{\link{interpolate.bfaov}}
# }
# \keyword{manip}
#*/

rg.remove.quantile <- function(RG,p.low=0.2,level=1,p.high=1,
    intensities=T,set.na=F) {
  fl <- rg.failures.quantile(RG,p.low,p.high,intensities=intensities)
  ind <- which(apply(fl,1,sum) >= level)
  if (level > 1 && set.na) {
    #fl[fl >= 1] <- T
    if (intensities) {
      # sum
      RG$R[fl] <- NA
      RG$Rb[fl] <- NA
      RG$G[fl] <- NA
      RG$Gb[fl] <- NA
    } else {
      # separate channels
      n <- ncol(fl)
      RG$R[fl[,1:(n/2)]] <- NA
      RG$Rb[fl[,1:(n/2)]] <- NA
      RG$G[fl[,(n/2+1):n]] <- NA
      RG$Gb[fl[,(n/2+1):n]] <- NA
    }
  }
  RG.rm <- rg.remove(RG,ind)
  RG.rm$inds <- RG$inds[-ind]
  RG.rm
}

a.remove.quantile <- function(a,p.low=0.2,level=1,p.high=1,
    set.na=F) {
  fl <- a.failures.quantile(a,p.low,p.high)
  ind <- which(apply(fl,1,sum) >= level)
  if (level > 1 && set.na) {
      # separate channels
      n <- ncol(fl)
      a[fl] <- NA
  }
  if (length(ind) > 0)
    a[-ind,]
  else
    a
}

rg.failures.quantile <- function(RG,p.low,p.high=1,intensities=T) {
  # separate channels R,G
  if (intensities) {
  # product of R, G:
    R <- RG$R-RG$Rb; G <- RG$G - RG$Gb
    R[R <= 0] <- 1; G[G <= 0] <- 1
    rg <- R*G
    q.low <- apply(rg,2,quantile,p.low,na.rm=T)
    q.high <- apply(rg,2,quantile,p.high,na.rm=T)
  } else {
    rg <- cbind(RG$R-RG$Rb,RG$G - RG$Gb)
    q.low <- apply(rg,2,quantile,p.low)
    q.high <- apply(rg,2,quantile,p.high)
  }
  ind <- apply(rg,1,function(v) v < q.low | v > q.high)
  t(ind)
}

a.failures.quantile <- function(a,p.low,p.high=1) {
  q.low <- apply(a,2,quantile,p.low)
  q.high <- apply(a,2,quantile,p.high)
  ind <- apply(a,1,function(v) v < q.low | v > q.high)
  t(ind)
}

#/**
# \name{rg2ma}
# \alias{rg2ma}
# \alias{ma2rg}
# \alias{ma.write}
# \title{Interconversion of RG and MA structure}
# \description{
#   Converts R, G channel data into M (differential expression),
#   A (intensity) data
# }
# \usage{
#   rg2ma(RG,normalize=F)
#   ma2rg(MA)
#   ma.write(RG,i=0,file="ma.dat",normalize=T)
# }
# @synopsis
# \arguments{
#   \item{RG}{RG, R, G channel microarray data structure,
#     see \code{\link{read.rg}}}
#   \item{MA}{MA, differential expression, intensity microarray
#     data structure}
#   \item{normalize}{if TRUE, \eqn{\log_2(R)}{log2(R)}
#     and \eqn{\log_2(G)}{log2(G)} values are
#     centralized}
#   \item{file}{file name}
#   \item{i}{index of experiment to write, if 0 average of A and M values
#     are used}
# }
# \details{
#   \code{MA$A} contains
#   \eqn{(log_2(R-Rb)+log_2(G-Gb))/2}{(log2(R-Rb)+log2(G-Gb))/2} and
#   \code{MA$M} contains
#   \eqn{(log_2(R-Rb)-log_2(G-Gb))/2}{(log2(R-Rb)-log2(G-Gb))/2}. This
#   amounts to a rotation by \eqn{\pi/4}{pi/4} of the
#   \eqn{log_2(R)}{log2(R)} over \eqn{log_2(G)}{log2(G)} plot.
#   \code{ma.write} writes the MA values to a file.
# }
# \value{
#   an RG or MA structure, compare \code{\link{init.data}}
#   from the \code{sma} package
# }
# \examples{
#   data(trcs)
#   MA <- rg2ma(trcs.RG)
#   plot(MA$A[,1],MA$M[,1])
#   RG <- ma2rg(MA)
# }
# \seealso{
#   \code{\link{read.rg}}, \code{\link{init.data}}
# }
# \keyword{manip}
#*/

rg2ma <- function(RG,normalize=F) {
  # a = lr + lg / 2, m = lr - lg
  # lr = a + m/2 , lg = a - m/2
  r <- RG$R - RG$Rb
  g <- RG$G - RG$Gb
  r [r <= 0] <- 1
  g [g <= 0] <- 1
  if (normalize) {
    lr <- log(r,2)
    lg <- log(g,2)
    M <- lr - mean(lr) - (lg - mean(lg))
  } else {
    M <- log(r,2) - log(g,2)
  }  
  A <- (log(r,2) + log(g,2))/2
  list(M=M,A=A,genes=RG$gene,experiments=RG$experiments,
      x=RG$x,y=RG$y)
}

ma2rg <- function(MA) {
  R <- 2^(MA$A + MA$M/2)
  Rb <- array(0,dim(R))
  G <- 2^(MA$A - MA$M/2)
  Gb <- array(0,dim(G))
  list(R=R,Rb=Rb,G=G,Gb=Gb,genes=MA$gene,experiments=MA$experiments,
      x=MA$x,y=MA$y)
}

ma.write <- function(RG,i=0,file="ma.dat",normalize=T) {
  MA <- rg2ma(RG,normalize=normalize)
  if (i == 0) {
    MA$A[,1] <- matrix(apply(MA$A,1,mean),,1)
    MA$M[,1] <- matrix(apply(MA$M,1,mean),,1)
    i <- 1
  }
  scale(MA$M,scale=F)
  dfr <- data.frame(number=1:length(MA$genes),name=MA$genes,logratio=MA$M[,i],
     logintensity=MA$A[,i] )
  write.table(dfr,file=file,sep="\t",
    quote=F,row.names=F)
}

#/**
# \name{rg.log.plot}
# \alias{rg.log.plot}
# \alias{ma.plot}
# \alias{ma.plot.single}
# \title{Plots log G over log R for RG structure}
# \description{
#   Plots logs of channel R over channel G
# }
# \usage{
#   rg.log.plot(RG,i=0,points=F,...)
#   ma.plot(RG,i=0,points=F,...)
#   ma.plot.single(RG,gene,i=0,points=T,...)
# }
# @synopsis
# \arguments{
#   \item{RG}{RG microarray data structure, see \code{\link{read.rg}}}
#   \item{i}{index of experiment to plot, if 0 an average over all is
#     shown}
#   \item{points}{if TRUE, \code{\link{plot}} is used, else points added
#     by \code{\link{points}}  }
#   \item{gene}{name of a single gene to plot}
#   \item{...}{arguments for \code{\link{plot}}}
# }
# \details{
#   A plot of \eqn{\log_2(R)}{log2(R)} over \eqn{\log_2(R)}{log2(R)} is
#   useful as a first diagnostic tool. \code{ma.plot} shows the plot
#   rotated by 45 degrees. If there is a
#   deviation from the diagonal a correction by
#   \code{\link{rg.loess.normalize}} or
#   \code{\link{rg.linear.normalize}} should be applied. 
# }
# \value{
#   no return value
# }
# \examples{
#   data(trcs)
#   ma.plot(trcs.RG,0,pch=".",xlab="intensities",ylab="diff expression") 
#   # shows the variability of a single gene across experiments:
#   ma.plot.single(trcs.RG,"dnaA",i=1:ncol(trcs.RG$R),
#       pch=19,col="blue",cex=0.5)
# }
# \seealso{
#   \code{\link{read.rg}}, \code{\link{init.data}},
#   \code{\link{rg2ma}}, \code{\link{ma2rg}}
# }
# \keyword{manip}
#*/

rg.log.plot <- function(RG,i=0,points=F,...) {
  r <- RG$R-RG$Rb
  g <- RG$G-RG$Gb
  r[r <= 0] <- 1
  g[g <= 0] <- 1
  lr <- log(r,2)
  lg <- log(g,2)
  if (i == 0) {
    lr <- matrix(apply(lr,1,mean),,1)
    lg <- matrix(apply(lg,1,mean),,1)
    i <- 1
  }
  if (!points)
    plot(lg[,i],lr[,i],...)
  else
    points(lg[,i],lr[,i],...)
}
  
ma.plot <- function(RG,i=0,points=F,...) {
  MA <- rg2ma(RG)
  if (i == 0) {
    MA$A[,1] <- matrix(apply(MA$A,1,mean),,1)
    MA$M[,1] <- matrix(apply(MA$M,1,mean),,1)
    i <- 1
  }
  if (!points)
    plot(MA$A[,i],MA$M[,i],...)
  else
    points(MA$A[,i],MA$M[,i],...)
}
  
ma.plot.single <- function(RG,gene,i=0,points=T,...) {
  g <- pmatch(gene,RG$genes)
  ma.plot(rg.keep(RG,g),i=i,points=points,...)
}

#/**
# \name{rg.loess.normalize}
# \alias{rg.loess.normalize}
# \alias{rg.linear.normalize}
# \alias{rg.shift.normalize}
# \alias{rg.scale.normalize}
# \alias{rg.rank.normalize}
# \alias{ma.cor}
# \title{Loess normalization of RG values}
# \description{
#   Corrects for dependency of log differential expression on
#   total log intensity by fitting a loess or linear regression.
# }
# \usage{
#   rg.loess.normalize(RG,deg=1,span=0.6,flat=F,
#       show=T,show.new=F,...)
#   rg.linear.normalize(RG,method=c("lm","prcomp"),
#       show=T,show.new=F,...)
#   rg.shift.normalize(RG,method=c("mean","median"))
#   rg.rank.normalize(RG,method=c("mean","median"))
#   rg.scale.normalize(RG,method=c("sd","mad"))
#   ma.cor(RG,i=0,show=T,...)
# }
# @synopsis
# \arguments{
#   \item{RG}{RG microarray data structure, see \code{\link{read.rg}}}
#   \item{deg}{degree of loess fit, the degree of the local
#     regression function, see \code{\link{loess}}}
#   \item{span}{span of loess fit, approximately the 
#     fraction of points involved in the local regression,
#     see \code{\link{loess}}}
#   \item{method}{"lm", linear regression; "prcomp", fit principal
#     component axes and rotate;
#     "mean", "median" mean and median normalization}
#   \item{scale}{if TRUE, scale to sd 1}
#   \item{show}{if TRUE, plots are shown}
#   \item{show.new}{if TRUE, add points after correction}
#   \item{i}{index of experiment to write, if 0 average of A and M values}
#   \item{flat}{if TRUE, plot a reference 0 line instead of loess curve}
#   \item{mfrow}{mfrow argument for panel layout, see \code{\link{par}}}
#   \item{pick}{a vector of array numbers to show}
#   \item{...}{parameters for function \code{\link{loess}}}
# }
# \details{
#   Converts an RG to an AM structure (A, x-axis log total intensity;
#   M, y-axis log difference between R and G), and normalizes AM plot by
#   fitting a loess regression (\code{rg.loess.normalize}) or linear
#   regression (\code{rg.linear.normalize}) and subtracting estimated
#   values. Plots of loess curves are shown. For more versatile
#   functions in this direction see \code{sma} package and
#   \code{\link{init.data}}. \code{ma.cor} plots and caculates the
#   correlation between A and M.
# }
# \value{
#   an RG structure with normalized R and G values
# }
# \examples{
# \dontrun{
#   data(trcs)
#   RG.norm <- rg.loess.normalize(trcs.RG,deg=2,span=0.4)
# }
# }
# \seealso{
#   \code{\link{read.rg}}, \code{\link{init.data}},
#   \code{\link{loess}},\code{\link{prcomp}}, \code{\link{lm}}}
# }
# \keyword{manip}
#*/

rg.loess.normalize <- function(RG,deg=1,span=0.6,show=T,
    show.new=F,flat=F,...) {
  MA <- rg2ma(RG)
  d <- ncol(MA$A)
  opar <- par(mfrow=c(round(sqrt(d)),ceiling(d/round(sqrt(d)))))
  on.exit(par(opar))
  for (i in 1:d) {
    loe <- loess(MA$M[,i]~MA$A[,i],deg=1,span=0.6)
    if (show)
      plot(MA$A[,i],MA$M[,i],pch=".",xlab="1/2 log(R*G)",ylab="log(R/G)",...)
    MA$M[,i] <- MA$M[,i] - predict(loe,MA$A[,i])
    r <- range(MA$A[,i])
    z <- seq(r[1],r[2],length=100)
    if (show) {
      if (!flat) {
        lines(z,predict(loe,z),col="red",lwd=2)
      }
      else  {
        lines(c(0,20),c(0,0),col="red",lwd=2)
      }
      if (show.new)
        points(MA$A[,i],MA$M[,i],pch=".",col="red")
    }
  }
  ma2rg(MA)
}

rg.linear.normalize <- function(RG,method=c("lm","prcomp"),
    show.new=F,show=T,mfrow=NULL,pick=NULL,...) {
  method <- match.arg(method)
  MA <- rg2ma(RG)
  d <- ncol(MA$A)
  if (is.null(pick))
    pick <- 1:d
  if (is.null(mfrow))
    opar <- par(mfrow=c(round(sqrt(d)),ceiling(d/round(sqrt(d)))))
  else
    opar <- par(mfrow=mfrow)
  on.exit(par(opar))
  for (i in pick) {
    if (show)
      plot(MA$A[,i],MA$M[,i],pch=".",xlab="1/2 log(R*G)",ylab="log(R/G)",...)
#      plot(MA$A[,i],MA$M[,i],pch=".",xlab="",ylab="")
    r <- range(MA$A[,i])
    z <- seq(r[1],r[2],length=100)
    if (method == "prcomp") {
      m <- cbind(MA$A[,i],MA$M[,i])
      m.mean <- apply(m,2,mean)
      mc <- scale(m,scale=F)  # centralize
      pc <- prcomp(m) # or the same: mc
      mcr <- mc %*% pc$rot  # rotate centralized matrix
      mcrt <- t(t(mcr) + c(m.mean[1],0))  # shift by x = mean(x), 0
      MA$A[,i] <- mcrt[,1]
      MA$M[,i] <- mcrt[,2]
      if (show)
        lines(z,(z-m.mean[1])*pc$rot[2,1]/pc$rot[1,1],col="red",lwd=1)
    } else {
      m <- lm(MA$M[,i]~MA$A[,i])
      MA$M[,i] <- MA$M[,i] - (m$coef[1] + m$coef[2]*MA$A[,i])
      if (show)
        lines(z,m$coef[1] + m$coef[2]*z,col="red",lwd=1)
    }
    if (show && show.new)
      points(MA$A[,i],MA$M[,i],pch=".",col="red")
  }
  ma2rg(MA)
}

rg.shift.normalize <- function(RG,method=c("mean","median"),do.log=T) {
  method <- match.arg(method)
  fct <- switch(method,median=median,mean=mean)
  MA <- rg2ma(RG)
  s <- apply(MA$M,2,fct)
  MA$M <- t(apply(MA$M,1,function(v) v - s))
  ma2rg(MA)
}

rg.scale.normalize <- function(RG,method=c("sd","mad",)) {
  method <- match.arg(method)
  fct <- switch(method,sd=sd,mad=mad)
  MA <- rg2ma(RG)
  s <- apply(MA$M,2,fct)
  MA$M <- t(apply(MA$M,1,function(v) v/s))
  ma2rg(MA)
}

rg.rank.normalize <- function(RG,method=c("mean","median")) {
  method <- match.arg(method)
  MA <- rg2ma(RG)
  MA$M <- rank.normalize(MA$M,method)
  ma2rg(MA)
}

rank.normalize <- function(ap,method=c("mean","median")) {
  method <- match.arg(method)
  o.ap <- apply(ap,2,order)
  ap.o <- NULL
  for (i in 1:ncol(o.ap))
    ap.o <- cbind(ap.o,ap[o.ap[,i],i])
  m.ap <- apply(ap.o,1,method)
  ap.n <- array(0,dim(ap))
  for (i in 1:ncol(ap.n))
    ap.n[o.ap[,i],i] <- m.ap
  ap.n
}

#rg.scale.normalize <- function(RG,method=c("mean","median"),
#    scale=F,pick=NULL) {
#  method <- match.arg(method)
#  MA <- rg2ma(RG)
#  d <- ncol(MA$A)
#  if (is.null(pick))
#    pick <- 1:d
#  for (i in pick) {
#    if (method == "mean") {
#      MA$M[,i] <- scale(MA$M[,i],scale=scale)
#    } else if (method == "median") {
#      MA$M[,i] <- MA$M[,i] - median(MA$M[,i])
#    }
#  }
#  ma2rg(MA)
#}

ma.cor <- function(RG,i=0,show=T,...) {
  MA <- rg2ma(RG)
  if (i == 0) {
    MA$A[,1] <- matrix(apply(MA$A,1,mean),,1)
    MA$M[,1] <- matrix(apply(MA$M,1,mean),,1)
    i <- 1
  }
  a <- MA$A[,i]
  m <- MA$M[,i]
  lm.ma <- lm(m~a)
  if (show) {
    plot(a,m,pch=".")
    abline(lm.ma)
  }
  cor.test(a,m) # cor is lm.ma$coef[2] * sd(a) / sd(m)
}
  
#/**
# @name
# \title{Loess normalization of RG values over microarray}
# \description{
#   Normalizes the log of R and G values separately (or their combined
#   log difference) over the microarray by subtracting a
#   two-dimensional loess local trend surface.
# }
# \usage{
#   rg.loess.array.normalize(RG,
#     method=c("ratio","rg","bg"),
#     deg=2,span=0.35,persp=F,
#     n.grid=100,n.color=200,show.color=T,
#     show.points=F,
#     auto.mfrow=T,p.high=1.0,p.low=0.0)
# }
# @synopsis
# \arguments{
#   \item{RG}{RG microarray data structure, see \code{\link{read.rg}},
#     with x (\code{RG$x}) and y (\code{RG$y}) coordinates}
#   \item{method}{"ratio", log ratio values are normalized;
#     "rg", \eqn{\log_2(R)}{log2(R)} and
#        \eqn{\log_2(R)}{log2(R)} are normalized separately,
#     "bg", only backround is normalized}
#   \item{deg}{degree of \code{\link{loess}} fit, default 2}
#   \item{span}{span of \code{\link{loess}} fit, default 0.35}
#   \item{persp}{if TRUE, perspective views on surfaces instead of contour
#     plots are shown}
#   \item{n.grid}{number of grid points in plots}
#   \item{n.color}{number of color shades}
#   \item{show.color}{if TRUE, color plots are shown, gray plots else}
#   \item{show.points}{if TRUE, gene spots are shown}
#   \item{p.low,p.high}{lower and upper percentiles of values to remove
#     before fitting surface}
#   \item{auto.mfrow}{if TRUE, layout of plots calculated}
# }
# \details{
#   Often uneven hybridization across the array is observed. This can
#   be corrected by fitting and subtracting a 2-dimensional loess
#   surface. It might be either applied to the log ratio values or to
#   the two channels separately (or to the background).
# }
# \value{
#   an RG structure with normalized R and G values, Rb and Gb are zero
# }
# \examples{
#   data(trcs)
#   RG.norm <- rg.loess.array.normalize(trcs.RG)
#   # R and R separately for array 1
#   tmp <- rg.loess.array.normalize(rg.project(trcs.RG,1),
#       method="rg", show.points=T)
# }
# \seealso{
#   \code{\link{read.rg}}, \code{\link{init.data}},
#   \code{\link{loess}},
# }
# \keyword{manip}
#*/

rg.loess.array.normalize <- function(RG,
    method=c("ratio","rg","bg"),
    deg=2,span=0.35,persp=F,
    n.grid=100,n.color=200,show.color=T,show.points=F,
    auto.mfrow=T,p.high=1.0,p.low=0.0) {
  method <- match.arg(method)
  dim <- 1:ncol(RG$R)
  d <- length(dim)
  if (method == "rg" || method == "bg") {
    RG$Rpred <- RG$R
    RG$Gpred <- RG$G
    d <- 2*d
    if (show.color == T) {
      #colors.r <- rgb(seq(0.5,1,len=n.color),0,0)
      #colors.g <- rgb(0,seq(0.5,1,len=n.color),0)
      colors.r <- rainbow(n.color,start=0,end=2/6)[n.color:1]
      colors.g <- rainbow(n.color,start=0,end=2/6)
    } else {
      colors.r <- grey(seq(0.2,1,len=n.color))
      colors.g <- colors.r
    }
  } else if (method == "ratio") {
    MA <- rg2ma(RG)
    pred <- RG$R
    if (show.color == T) {
      colors <- rainbow(n.color,start=0,end=2/6)[n.color:1]
    } else {
      colors <- grey(seq(0.2,1,len=n.color))
    }
  }
  if (auto.mfrow) {
    opar <- par(mfrow=c(round(sqrt(d)),ceiling(d/round(sqrt(d)))))
    on.exit(par(opar))
  }
  show.plot <- (n.color > 0)
  for (i in dim) {
    if (method == "rg" || method == "bg") {
      if (method == "rg") {
        dr <- RG$R[,i] - RG$Rb[,i]
        dg <- RG$G[,i] - RG$Gb[,i]
      }
      else {
        dr <- RG$Rb[,i]
        dg <- RG$Gb[,i]
      }
      dr[dr <= 0] <- 1
      dg[dg <= 0] <- 1
      x <- RG$x[,i]
      y <- RG$y[,i]
      z <- log(dr,2)
      RG$Rpred[,i] <- loess.2d(x,y,z,deg=deg,span=span,persp=persp,
              n.grid,colors=colors.r,show.points,show.plot,
              p.high=p.high,p.low=p.low)
      z <- log(dg,2)
      RG$Gpred[,i] <- loess.2d(x,y,z,deg=deg,span=span,persp=persp,
              n.grid,colors=colors.g,show.points,show.plot,
              p.high=p.high,p.low=p.low)
      if (method == "rg") {
        RG$R[,i] <- exp(z - RG$Rpred[,i])
        RG$G[,i] <- exp(z - RG$Gpred[,i])
        RG$Rb[,i] <- 0
        RG$Gb[,i] <- 0
      }
      else {
        RG$Rb[,i] <- exp(z - RG$Rpred[,i])
        RG$Gb[,i] <- exp(z - RG$Gpred[,i])
      }
    } else if (method == "ratio") {
      x <- RG$x[,i]
      y <- RG$y[,i]
      z <- MA$M[,i]
      pred[,i] <- loess.2d(x,y,z,deg=deg,span=span,persp=persp,
              n.grid,colors=colors,show.points,show.plot,
              p.high=p.high,p.low=p.low)
      MA$M[,i] <- z - pred[,i]
    }
  }
  if (method == "ratio") {
    RG <- ma2rg(MA)
    RG$pred <- pred
    RG
  }
  RG
}

loess.2d <- function(x,y,z,deg,span,persp,
      n.grid,show.points,show.plot=T,colors=NULL,
      p.low,p.high) {
  q.low <- quantile(z,p.low)
  q.high <- quantile(z,p.high)
  inds <- intersect(which(z > q.low),which(z < q.high))
  # loess can only interpolate within defining range:
  x.min <- which(x == min(x))
  x.max <- which(x == max(x))
  y.min <- which(y == min(y))
  y.max <- which(y == max(y))
  inds <- c(inds,unique(c(x.min,x.max,y.min,y.max)))
  zl <- z[inds]; xl <- x[inds]; yl <- y[inds]
  ma.loess <- loess(zl ~ xl * yl,degree=deg,span=span,normalize=F)
    #control=loess.control(surface="interpolate") )
  if (show.plot) {
    if (!persp) {
      xx <- seq(min(x),max(x),len=n.grid)
      yy <- seq(min(y),max(y),len=n.grid)
    } else {
      xx <- seq(min(x),max(x),len=30)
      yy <- seq(min(y),max(y),len=30)
    }  
    ma.plt <- expand.grid(xl=xx,yl=yy)

    # normal plot
    zz <- matrix(predict(ma.loess,ma.plt),length(xx))
    # error plot:
    #zz <- matrix(predict(ma.loess,ma.plt,se=T)$se,length(xx))

    if (is.null(colors))
      colors <- rainbow(200,start=0,end=2/6)[200:1]

    if (!persp) {
      image(xx,yy,zz,col=colors,axes=F,xlab="",ylab="")
      ma.surf <- list(x=xx,y=yy,z=zz)
      contour(ma.surf,labcex=1.2,add=T,nlevels=20,axes=F,xlab="",ylab="")
      if (show.points)
        points(x,y,pch=".")
    } else {
      persp(xx,yy,zz,theta=-50,phi=20,expand=0.5,col="lightblue",
          ltheta=120,shade=0.75,
          xlab="array x",ylab="array y",zlab="log(R/G)")
    }
  }

  ma.pre <- data.frame(xl=x,yl=y)
  as.vector(predict(ma.loess,ma.pre))
}

#/**
# \name{rg2array}
# \alias{rg2array}
# \alias{rg2log.array}
# \alias{array2rg}
# \alias{log.array2rg}
# \title{Balanced factorial array from RG structure}
# \description{
#   Generates a balanced factorial array suitable for the
#   ANOVA analysis for balanced factorial designs.
# }
# \usage{
#   rg2array(RG,experiments=RG$experiments)
#   rg2log.array(RG,experiments=RG$experiments)
#   array2rg(a)
#   log.array2rg(a)
# }
# @synopsis
# \arguments{
#   \item{RG}{RG microarray data structure, see \code{\link{read.rg}}}
#   \item{experiments}{see \code{\link{read.rg}} for a description of
#   the experiment list}
#   \item{a}{a data array according to a balanced factorial design}
# }
# \details{
#   Generates a balanced factorial array suitable for the
#   \code{\link{bfaov}} ANOVA analysis for balanced factorial
#   designs. Such bfaov array is a more efficient way of storing data
#   from a balanced factorial design than a data frame.
#   An array is constructed with each dimension corresponding to a
#   factor as descripted in \code{experiments}. Two dimensions are
#   added, one for the gene G effect and one for the variety V effect
#   (background corrected red vs background corrected
#   green). \code{rg2array} array does \emph{not} transformed the data
#   by any further operation (eg log). \code{rg2log.array} transforms
#   by \eqn{\log_2}{log_2} (log of nonpositive values set to
#   0). {\code{names(dim(a))}} contains the names of the dimensions.
# }
# \value{
#   a multidimensional array
# }
# \examples{
#   data(trcs)
#   a <- rg2log.array(trcs.RG)
#   av <- bfaov(~(V+G)^2,a)
#   anova(av)
# }
# \seealso{
#   \code{\link{read.rg}}, \code{\link{bfaov}}
# }
# \keyword{manip}
#*/

rg2array <- function(RG,experiments=RG$experiments) {
  # Construct data array from RG data
  # a dimension for each variable

  if (data.class(experiments) != "data.frame") {
    experiments <- rev(do.call("expand.grid",rev(experiments)))
  }
  dims <- sapply(experiments, function(x) length(unique(x)))
  g <- dim(RG$R)[1]
  # mod. P. Sykacek 2008 - latest R breaks original code!!
  dims <- c(2, g, dims)
  names(dims)[1:2] <- c("V","G")
  a <- array(0,dims)
  names(dim(a))<- names(dims)
  for (i in 1:nrow(experiments)) {
    d <- experiments[i,]
    cmd1 <- paste("a[1,,",paste(d,sep="",collapse=","),
      "] <- RG$R[,i]-RG$Rb[,i]",sep="",collapse="") 
    eval(parse(text=cmd1))
    cmd2 <- paste("a[2,,",paste(d,sep="",collapse=","),
      "] <- RG$G[,i]-RG$Gb[,i]",sep="",collapse="") 
    eval(parse(text=cmd2))
  }
  attr(a,"genes") <- RG$genes
  a
}

rg2log.array <- function(RG,log.ratio=F,log.intensity=F,
    experiments=RG$experiments) {
  a <- rg2array(RG,experiments)
  a[a <= 0] <- 1
  al <- log(a,2)
  if (log.ratio || log.intensity) {
    names <- names(dim(al))
    rest.dims <- dim(al)[-1]
    al.r <- array(al,c(2,prod(rest.dims)))
    if (log.intensity) {
      al <- array(0.5*(al.r[1,] + al.r[2,]),rest.dims)
    } else {
      al <- array(al.r[1,] - al.r[2,],rest.dims)
    }
    names(dim(al)) <- names[-1]
  }
  al
}

array2rg <- function(a) {
  # Construct RG data from array

  experiments <- list()
  names <- names(dim(a))
  inds <- seq(along=names)[-match(c("V","G"),names)]
  for (i in inds)
    experiments <- c(experiments,list(1:dim(a)[i]))
  names(experiments) <- names[inds]

  exp.dfr <- rev(do.call("expand.grid",rev(experiments)))
  dims <- sapply(exp.dfr, function(x) length(unique(x)))
  g <- dim(a)["G"]
  rdims <- prod(dim(a)[inds])
  RG <- list()
  RG$R <- matrix(0,g,rdims);
  RG$Rb <- matrix(0,g,rdims)
  RG$G <- matrix(0,g,rdims);
  RG$Gb <- matrix(0,g,rdims)
  for (i in 1:nrow(exp.dfr)) {
    d <- exp.dfr[i,]
    cmd1 <- paste("RG$R[,i] <- ",
      "a[1,,",paste(d,sep="",collapse=","),"]",sep="",collapse="") 
    eval(parse(text=cmd1))
    cmd2 <- paste("RG$G[,i] <- ",
      "a[2,,",paste(d,sep="",collapse=","),"]",sep="",collapse="") 
    eval(parse(text=cmd2))
  }
  RG$genes <- attr(a,"genes")
  RG$experiments <- experiments
  RG
}

log.array2rg <- function(a) {
  a <- 2^a
  array2rg(a)
}

rg.log.avg <- function(RG) {
	RGa$R <- as.matrix(exp(apply(log(RG$R - RG$Rb,2),1,mean)))
	RGa$G <- as.matrix(exp(apply(log(RG$G - RG$Gb,2),1,mean)))
	RGa$Rb <- 0
	RGa$Gb <- 0
  RGa
}

#stat.Newton(RG,lay=set.grid(4,4,17,17),imag=2,norm="l",pout=T)


# QUALITY DIAGNOSTICS

#/**
# @name
# \title{Plot of number of low expression genes}
# \description{Plots the number of genes that are in the lower (upper)
#   quantiles of a given number of R or G channels}
# @usage
# @synopsis
# \arguments{
#   \item{RG}{RG microarray data structure, see \code{\link{read.rg}}}
#   \item{p.low}{the lower quantile to remove from all channels}
#   \item{p.high}{the upper quantile to remove from all channels}
#   \item{intensities}{if TRUE, spot intensities (R*G), otherwise
#     separate R, G channel values are used}
# }
# \details{
#   The value of a gene may fall in the lower (upper) \code{p.low}
#   (\code{p.high}) intensity quantile of one, two, or more arrays of
#   the experiments. The number of genes with values in the quantiles
#   of one, two, etc. arrays is shown in the histogram.
# }
# \value{
#   vector \code{v}: \code{v[k]} is the number of genes
#   which are in the lower (and upper) quantile of exactly
#   \code{k-1} arrays
# }
# \examples{
#   data(trcs)
#   v <- rg.failure.histogram(trcs.RG,0.1)
#   v
#   RG <- rg.remove.quantile(trcs.RG,0.1,level=3,set.na=T)
#   # number of genes with low values on one or on two arrays:
#   v[2] + v[3]
#   # number of missing values for this genes:
#   sum(is.na(RG$R)) == v[2] + 2*v[3]
# }
# \seealso{
#   \code{\link{read.rg}}, \code{\link{rg.remove.quantile}},
#   \code{\link{interpolate.bfaov}}
# }
# \keyword{hplot}
#*/

rg.failure.histogram <- function(RG,p.low,p.high=1,intensities=T) {
  fl <- rg.failures.quantile(RG,p.low,p.high,intensities=intensities)
  ft <- table(apply(fl,1,sum))
  if (intensities)
    plot(seq(along=ft[-1]),ft[-1],ty="h",
        xlab="number arrays with failure for the same gene",
        ylab="number genes")
  else
    plot(seq(along=ft[-1]),ft[-1],ty="h",
        xlab="number channels with failure for the same gene",
        ylab="number genes")
  ft
}

a.failure.histogram <- function(a,p.low=0.1,p.high=1.0) {
  fl <- a.failures.quantile(a,p.low,p.high)
  ft <- table(apply(fl,1,sum))
  plot(seq(along=ft[-1]),ft[-1],ty="h",
      xlab="number channels with failure for the same gene",
      ylab="number genes")
  ft
}

#/**
# \name{rg.rsq.plot}
# \alias{rg.rsq.plot}
# \alias{rg.rsq}
# \title{Diagnostics of low expression genes}
# \description{Plots correlation \eqn{R^2}{R^2} between
#   all experiments over lower quantiles of data points removed}
# \usage{
#   rg.rsq.plot(RG,p.end=0.6,p.start=0.0,
#       method=c("spearman","pearson"),
#       length=20,level=1,p.high=1,
#       do.prcomp=F,intensities=T)
#   rg.rsq(RG,p.low=0.0,
#       method=c("spearman","pearson"),
#       level=1,p.high=1,
#       do.prcomp=F,intensities=T)
# }
# \arguments{
#   \item{RG}{RG microarray data structure}
#   \item{p.start, p.end}{the range of lower quantiles of points to
#     remove}
#   \item{method}{"spearman", Spearman rank correlation;
#     "pearson", Pearson correlation}
#   \item{length}{number of evaluation points of plot}
#   \item{level}{number of experiments for which a data point must lie in
#     the given quantiles to be removed,
#     compare \code{\link{rg.failure.histogram}}}
#   \item{p.high}{the upper quantile to remove from all experiments}
#   \item{p.low}{the lower quantile to remove before calculating 
#         \eqn{R^2}{R^2}}
#   \item{do.prcomp}{see \code{\link{rg.remove.quantile}}}
#   \item{intensities}{see \code{\link{rg.remove.quantile}}}
# }
# \details{
#   How effective is \code{\link{rg.remove.quantile}} in improving the
#   quality of genes not removed? The overall \eqn{R^2}{R^2} after
#   removal of lowest quantiles of intensity genes is caculated for
#   various quantiles. Log ratio values \eqn{y_{gk}}{y_gk} are
#   considered for gene \eqn{g}{g} and array \eqn{k}{k}. The
#   \eqn{R^2}{R^2} of a one-way ANOVA with genes as treatment groups
#   is calculated, the ratio of variation between gene variation
#   (estimated by averages) and the total variation. A variant uses
#   variation along the principal axis compared to total variation.
#   The more data points with low spot intensity and high variation
#   are removed the better the signal usually becomes. This is shown
#   in the plot. To remove data points, choose a suitable quantile
#   resulting in a good signal without removing too many points
#   (usually around 0.1 to 0.3).
# }
# \value{
#   \code{rg.rsq} returns \eqn{R^2}{R^2} value
# }
# \examples{
#   data(trcs)
#   rg.rsq.plot(trcs.RG)
# }
# \seealso{
#   \code{\link{rg.remove.quantile}}
# }
# \keyword{hplot}
#*/

rg.rsq.plot <- function(RG,p.end=0.6,p.start=0.0,
  method=c("spearman","pearson"),length=20,level=1,p.high=1,
  do.prcomp=F,intensities=T) {
  l <- rg.log.matrix(RG)
  method <- match.arg(method)
  if (method == "spearman")
    l <- apply(l,2,rank)
  y <- NULL
  z <- seq(p.start,p.end,length=length)
  for (p.low in z) {
    #fl <- rg.failures.quantile(RG,p.low,p.high,intensities=intensities)
    #ind <- which(apply(fl,1,sum) >= level)
    #lf <- if (length(ind) > 0) l[-ind,] else l    
    #pc <- prcomp(lf,retx=T,scale=F)
    #y <- c(y,(pc$sdev[1])^2/sum(pc$sdev^2))
    y <- c(y,rg.rsq(RG,p.low=p.low,method=method,level=level,
      do.prcomp=do.prcomp,intensities=intensities))
  }
  plot(z,y,pch="+",ty="p",xlab="fraction",ylab="R^2")
  lines(ksmooth(z,y,"normal",bandwidth=1.5*(p.end-p.start)/length),
      ty="l",col="red")  
}

a.rsq.plot <- function(l,p.end=0.6,p.start=0.0,
  method=c("spearman","pearson"),length=20,level=1,p.high=1,
  do.prcomp=F) {
  method <- match.arg(method)
  if (method == "spearman")
    l <- apply(l,2,rank)
  y <- NULL
  z <- seq(p.start,p.end,length=length)
  for (p.low in z) {
    y <- c(y,a.rsq(l,p.low=p.low,method=method,level=level,
      do.prcomp=do.prcomp) )
  }
  plot(z,y,pch="+",ty="p",xlab="fraction",ylab="R^2")
  lines(ksmooth(z,y,"normal",bandwidth=1.5*(p.end-p.start)/length),
      ty="l",col="red")  
}

rg.rsq <- function(RG,p.low=0.0,method=c("spearman","pearson",),
    level=1,p.high=1,do.prcomp=F,intensities=T) {
  l <- rg.log.matrix(RG)
  fl <- rg.failures.quantile(RG,p.low,p.high,intensities=intensities)
  a.fl.rsq(l,fl,p.low=p.low,method=method,level=level,
    p.high=p.high,do.prcomp=do.prcomp)
}

a.rsq <- function(a,p.low=0.0,method=c("spearman","pearson",),
    level=1,p.high=1,do.prcomp=F) {
  fl <- a.failures.quantile(a,p.low,p.high)
  a.fl.rsq(a,fl,p.low=p.low,method=method,level=level,
    p.high=p.high,do.prcomp=do.prcomp)
}

a.fl.rsq <- function(l,fl,p.low=0.0,method=c("spearman","pearson",),
    level=1,p.high=1,do.prcomp=F) {
  method <- match.arg(method)
  if (method == "spearman")
    l <- apply(l,2,rank)
  ind <- which(apply(fl,1,sum) >= level)
  if (length(ind) > 0)
    l <- l[-ind,]
  if (do.prcomp) {
    pc <- prcomp(l,retx=T,scale=F)
    R2 <- (pc$sdev[1])^2/sum(pc$sdev^2)
  } else {
    lc <- scale(l,scale=F)
    ls <- apply(lc,1,mean)
    lcv <- as.vector(lc)
    #R2 <- (sum(ls^2)/(length(ls)-ncol(l))) /
    #    (sum(lcv^2)/(length(lcv)-1))
    R2 <- ncol(l)*sum(ls^2) / sum(lcv^2)
  }
  R2
}

#/**
# \name{rg.cor}
# \alias{rg.cor}
# \title{Correlation between experiments}
# \description{
#   The pairwise correlation of log ratios of the experiments
# }
# \usage{
#   rg.cor(RG,method=c("pearson","kendall","spearman"),
#       cluster=c("correlation","euclidean"),
#       ptype=c("none","tex","txt"),file="",
#       show=T,avg=NULL,
#       clustering.method=c("ward","single","average","complete"))
# }
# @synopsis
# \arguments{
#   \item{RG}{RG microarray data structure}
#   \item{method}{arguments for the correlation function,
#     \code{"pearson"} is the standard rank correlation,
#     \code{"spearman"} and \code{"kendall"} are Spearman's and
#     Kendall's rank correlation coefficients,
#     note that \code{"kendall"} might be slow}
#   \item{cluster}{"correlation", (default) clustering distance is
#     \eqn{1-c^2}{1-c^2}; "euclidean", euclidean distance}
#   \item{ptype}{"tex", the printout is as a \emph{latex} table;
#     "txt", a tab delimited table}
#   \item{file}{output file for \code{ptype} "tex" and "txt"}
#   \item{show}{plot clustering tree, default T}
#   \item{avg}{a list of vectors indicating which experiments should
#     be averaged over to reduce the size of the correlation table}
#   \item{clustering.method}{clustering methods, see \code{\link{hclust}}}
# }
# \details{
#   The assumption is that the experiments are repetitions of the same
#   basic experiment and hence they should be very similar. How
#   similar they actually are can be tested with this function. It
#   also plots a hierarchical clustering tree according to either the
#   correlation matrix, or the euclidean distance of the experiments.
# }
# \value{
#   no return value
# }
# \examples{
#   data(trcs)
#   rg.cor(trcs.RG)
#   # average over spot quantifications, reducing 12 to 6 sets:
#   tmp <- rg.cor(trcs.RG,ptype="txt",
#       avg=list(1:2,3:4,5:6,7:8,9:10,11:12))
# }
# \keyword{cluster}
#*/

rg.cor <- function(RG,
    method = c("spearman","pearson", "kendall"),
    cluster=c("correlation","euclidean"),
    ptype=c("none","tex","txt"),file="",show=T,avg=NULL,
    clustering.method=c("ward","single","average","complete")) {

  a.cor(rg.log.matrix(RG),method=method,cluster=cluster,
    ptype=ptype,file=file,show=show,avg=avg,
    clustering.method=clustering.method)
}

a.cor <- function(al,
    method = c("spearman","pearson", "kendall"),
    cluster=c("correlation","euclidean"),
    ptype=c("none","tex","txt"),file="",show=T,avg=NULL,
    clustering.method=c("ward","single","average","complete") ) {
  opt.old <- options(warn=-1)
  on.exit(options(opt.old))
  method <- match.arg(method)
  clustering.method <- match.arg(clustering.method)
  if (method == "spearman") {
    al <- apply(al,2,rank)
  }    
  if (!is.null(avg)) {
    # reduce complex correlations trees by merging some
    # experiments as described in avg:
    alnew <- NULL
    for (i in seq(along=avg))
      alnew <- cbind(alnew,apply(al[,avg[[i]]],1,mean))
    al <- alnew
  }
  ptype <- match.arg(ptype)
  cluster <- match.arg(cluster)
  n <- ncol(al)
  cm <- matrix(0,n,n)
  if (method == "pearson" || method == "spearman") {
    for (i in 1:n) {
      for (j in 1:i) {
        cm[i,j] <- cor(al[,i],al[,j])
      }
    }
  }
  else if (method == "kendall") {
    for (i in 1:n) {
      for (j in 1:i) {
        cm[i,j] <- cor.coef(al[,i],al[,j],"kendall")
      }
    }
  }    
  if (ptype != "none") {
    switch (ptype,
      txt = print.txt.matrix(cm,digits=2,width=7,file=file,
          lower.triangle=T),
      tex = print.tex.matrix(cm,digits=2,width=7,file=file,
          tadjust=".",lower.triangle=T))
  }
  if (ncol(al) > 2 && show) {n
    if (cluster == "correlation") {
      rc <- cm + t(cm)
      diag(rc) <- 1
      d <- as.dist(1-rc^2)
      #main <- "Clustering of experiments for 1 - corr^2"
      main <- ""
    } else if (cluster == "euclidean") {
      d <- dist(t(al))
      #main <- "Clustering of experiments for euclidean"
      main <- ""
    }
    if (show)
      plot(hclust(d,method=clustering.method),
        sub="",xlab="",ylab=" ",main=main)
  }
  cm
}

cor.coef <- function(x,y,method=c("pearson", "kendall", "spearman")) {
  switch(match.arg(method),
    "pearson" = cor(x,y),
    "spearman" = cor(rank(x),rank(y)),
    "kendall"= .C("kendall_tau",
			   as.integer(length(x)),
			   as.double(x),
			   as.double(y),
			   tau = as.double(0),
			   PACKAGE = "ctest")$tau
  )
}



