/**
 * C routines for fast
 *  - t statistics bootstrap, t_boots
 *  - VG effect bootstrap,    vg_boots
 *  - interpolation coefficients, norm_coefs
 */

#include<stdlib.h>
#include<math.h>
#include<R.h>
#include <Rdefines.h>
#include <R_ext/Rdynload.h>

/*
the following patch was added in May 2005 by P. Sykacek:
To make yasma compatible with R-win32, I change drand48
to the internal R random number generator. drand48 is
still used outside mingw32. (which is the suggsted 
library to port to win32 for portability - the original code required specific
cygwin installations and failed in particular to run with the 
latest cygwin.x implementations.)
Hence this yasma.c uses my_rand() where the original used drand48().
This version now uses the R internal random numbers.
In addition I added all dynamic load stuff manually at the end since I could not get
my test call to my_rand executed from within R.
*/



double my_rand() 
// portability of random numbers (drand48 is not strict ANSI 
// and not guaranteed to be available in stdlib.h!)
{
#ifndef __MINGW32__
  double d_res=drand48();
#else
  double d_res=unif_rand();
#endif
  return d_res;
}

void rand_interface(double* ret)
{
  *ret=my_rand();
}

void t_boots(
    int *cycles,
    int *n,
    int *m,
    int *step_down,
    double *s,
    double *tf,
    int *ind,
    int *count) {

  int round;
  double *tf_end = tf + (*m)*(*n);
  double *pf, *ps;
  double m1, m2;
  int *pcount;
  int i, *pind, *ind_end;
  double max, t, x1,s1,x2,s2;

  ind_end = ind + *m;
  for (round = 0; round < *cycles; ++round) {
    /* printf("%d ",round); fflush(0); */

    for (pind = ind; pind < ind_end; ++pind) {
       *pind = (int)(my_rand() * *m);
    }
    max = -1e20;

    for (pf = tf, pcount = count, ps = s;
        pf < tf_end;
        pf += (*m), ++pcount, ++ps) {
      /* for each of n genes */
      m1 = m2 = 0;
      s1 = s2 = 0;
      for (pind = ind; pind < ind_end; ++pind) {
        x1 = *(pf + *pind);
        m1 += x1;
        s1 += x1*x1;
      }
      m1 /= *m;
      /* t = fabs(m1)/sqrt(s1 - m1*m1*(*m)); */
      t = m1/sqrt(s1 - m1*m1*(*m));
      /* t = m1; */
      if (*step_down) {
        if (t > max)
          max = t;
        if (max >= *ps)
          ++(*pcount);
      } else {
        if (t >= *ps)
          ++(*pcount);
      }
    }
  }
  /* printf("\n"); */
}

/**
 * matrix
 *       [,1] [,2] [,3] [,4] [,5]
 * [1,]    1    4    7   10   13
 * [2,]    2    5    8   11   14
 * [3,]    3    6    9   12   15
 *
 *       [,1] [,2] [,3] [,4] [,5]
 * [1,]   16   19   22   25   28
 * [2,]   17   20   23   26   29
 * [3,]   18   21   24   27   30
 *
 * first dim is dr = 3 (rest in case of more dimensions than 3)
 * last 2 dimensions are dg = dim(gdim)=5 and dv = dim(vdim)=2
 *
 * v1 effect is sum over first 3x5 layer, ie 0..dr*dg-1
 * v2 effect second layer, drg ... 2*dr*dg-1
 * 
 * vg1 effect for each g is 0 + g*dr ... 0 + (g+1)*dr-1
 * vg2 effect for each g is dr*dg + g*dr ... dr*dg + (g+1)*dr-1
 *
 */


void vg_boots(
    int *n,      /* bootstrap rounds */
    int *dr, int *dg, int *dv,    /* see above */
    double *e,   /* expectation matrix */
    double *dvg, /* the VG differences */
    double *r_smooth, double *r_scaled, /* residuals matrices */
    double *r,   /* storage for combined residuals */
    int *stats_only, /* only summary statistics collected */
    double *b    /* return (dg x n) return matrix of bootstrap results */
  ) {

  int i;
  double *p1, *p2;
  double *pg1, *pg2, *pse, *g_end;
  double *r_end, *r_vend, r_v, r_vg;
  double *p_dvg;
  double *pro, *prs, *pr, *pe;
  double *pb, *pbb, diff;
  double *pbb_over;

  int drg = *dr * *dg;
  int sel;
  double drgv = *dr * *dg * *dv;
  int gg, j;

  r_vend = r + drg;
  r_end = r + (int)drgv;

  for (i = 0, pb = b; i < *n; ++i) {
    if (!(i % 100)) {
      printf("%d ",i); fflush(0);
    }

    /* random shuffle r */
    for (pro = r_smooth, pr = r; pr < r_end; ++pro, ++pr) {
      *pr = *pro * *(r_scaled + (int)(my_rand() * drgv));
    }

    /* the simulated data values */
    for (pr = r, pe = e; pr < r_end; ++pr, ++pe) {
      *pr += *pe;
    }


    /* average V1 - V2 over all genes and rest */
    r_v = 0;
    for (p1 = r, p2 = r_vend; p1 < r_vend; ++p1, ++p2)
      r_v += *p1 - *p2;
    r_v /= drg;

    gg = 0;
    if (*stats_only) {
      pb = b;
      pbb = b + *dg;
      pbb_over = b + 2*(*dg);
      p_dvg = dvg;
      for (pg1 = r, pg2 = r_vend, j = 0;
          pg1 < r_vend;
          pg1 += *dr, pg2 += *dr, ++j,
          ++pb, ++pbb, ++pbb_over, ++p_dvg) {
        /* printf("  g %d\n",gg++); */
        g_end = pg1 + *dr;
        r_vg = 0;
        /* average V1 - V2 over rest for gene gg */
        for (p1 = pg1, p2 = pg2; p1 < g_end; ++p1, ++p2)
          r_vg += *p1 - *p2;
        r_vg /= *dr;
        diff = r_vg - r_v;
        *pb += diff;
        *pbb += diff*diff;
        /* if (j == 3376) {
           printf(" %lf/%lf ",diff,*p_dvg); fflush(0);
        } */
        if (diff < 0)
          (*pbb_over)++;
      }
    }
    else
      for (pg1 = r, pg2 = r_vend; pg1 < r_vend;
          pg1 += *dr, pg2 += *dr, ++pb) {
        /* printf("  g %d\n",gg++); */
        g_end = pg1 + *dr;
        r_vg = 0;
        for (p1 = pg1, p2 = pg2; p1 < g_end; ++p1, ++p2)
          r_vg += *p1 - *p2;
        r_vg /= *dr;
        *pb = r_vg - r_v;
      }

  }
}

/**
 * matrix
 *       [,1] [,2] [,3] [,4] [,5]
 * [1,]    1    4    7   10   13
 * [2,]    2    5    8   11   14
 * [3,]    3    6    9   12   15
 *
 * first dim is dr = 3 (rest in case of more dimensions than 3)
 * last dimension is dg = dim(gdim)=5
 *
 * mean(g) effect is sum over 3x5 layer, ie 0..dr*dg-1
 * g effect for each g is 0 + g*dr ... 0 + (g+1)*dr-1
 *
 */


void g_boots(
    int *n,      /* bootstrap rounds */
    int *dr,
    int *dg,    /* see above */
    double *e,   /* expectation matrix */
    double *r_smooth,
    double *r_scaled, /* residuals matrices */
    double *r,   /* storage for combined residuals */
    int *stats_only, /* only summary statistics collected */
    double *b    /* return (dg x n) return matrix of bootstrap results */
  ) {

  int i;
  double *p1;
  double *pg1, *pse, *g_end;
  double *r_end, r_mean;
  double *p_g, r_g;
  double *pro, *prs, *pr, *pe;
  double *pb, *pbb, diff;
  double *pbb_over;

  int drg = *dr * *dg;
  int gg, j;

  r_end = r + (int)drg;

  for (i = 0, pb = b; i < *n; ++i) {  /* foreach ROUND */
    if (!(i % 100)) {
      printf("%d ",i); fflush(0);
    }

    /* random shuffle r */
    for (pro = r_smooth, pr = r; pr < r_end; ++pro, ++pr) {
      *pr = *pro * *(r_scaled + (int)(my_rand() * drg));
    }

    /* the simulated data values */
    for (pr = r, pe = e; pr < r_end; ++pr, ++pe) {
      *pr += *pe;
    }

    /* average G over all genes and rest */
    r_mean = 0;
    for (p1 = r; p1 < r_end; ++p1)
      r_mean += *p1;
    r_mean /= drg;

    gg = 0;
    if (*stats_only) {
      pb = b;
      pbb = b + *dg;
      pbb_over = b + 2*(*dg);
      /* p_g = g; */
      for (pg1 = r, j = 0;           /* foreach GENE */
          pg1 < r_end;
          pg1 += *dr, ++j,
          ++pb, ++pbb, ++pbb_over) {
        /* printf("  g %d\n",gg++); */
        g_end = pg1 + *dr;
        r_g = 0;
        for (p1 = pg1; p1 < g_end; ++p1)
          r_g += *p1;
        r_g /= *dr;
        /* gene effect is gene average - total mean mu */
        diff = r_g - r_mean;
        *pb += diff;
        *pbb += diff*diff;
        /* if (j == 3376) {
           printf(" %lf ",diff); fflush(0);
        } */
        if (diff < 0)
          (*pbb_over)++;
      }
    }
    else
      for (pg1 = r; pg1 < r_end;
          pg1 += *dr, ++pb) {
        /* printf("  g %d\n",gg++); */
        g_end = pg1 + *dr;
        r_g = 0;
        for (p1 = pg1; p1 < g_end; ++p1)
          r_g += *p1;
        r_g /= *dr;
        *pb = r_g - r_mean;
      }

  }
}

/**
 * 1st matrix
 *          genes
 *      [,1] [,2] [,3]
 * [1,]    1    5    9
 * [2,]    2    6   10
 *
 * [3,]    3    7   11
 * [4,]    4    8   12
 *
 * 2nd matrix
 *          genes
 *      [,1] [,2] [,3]
 * [1,]   13   17   21
 * [2,]   14   18   22
 *
 * [3,]   15   19   23
 * [4,]   16   20   24
 *
 * first dim is dr = 4 (rest in case of more dimensions than 3)
 * last dimension is dg = dim(gdim)=3
 *
 * mean(g) effect is sum over 4x3layer, ie 0..dr*dg-1
 * g effect for each g is 0 + g*dr ... 0 + (g+1)*dr-1
 *
 *
 * Residues for higher levels are simply repeated:
 * 2nd matrix, residues for a higher level:
 *          genes
 *      [,1] [,2] [,3]
 * [1,]   5.8  9.3   10.5
 * [2,]   5.8  9.3   10.5
 *
 * [3,]   0.1  8.8   7.3
 * [4,]   0.1  8.8   7.3
 * 
 * Shuffling takes place within each stratum (eg here 
 * with span[2] == 2)
 *
 */

void g_level_boots(
    int *n,      /* bootstrap rounds */
    int *dh,      /* number of levels */
    int *dr,
    int *dg,    /* see above */
    double *e_matrix,   /* expectation matrix */
    double *a_random,  /* room for shuffled values */
    double *r_smooth,
    double *r_scaled, /* residuals matrices */
    double *gene_limit, /* gene estimate */
    int *spans,   /* the span in each level */
    int *stats_only, /* only summary statistics collected */
    double *b  /* return (dg x n) return matrix of bootstrap results */
  ) {

  int i, gg, j, h;
  double *p1;
  double *pg1, *pse, *g_end;
  double *r_end, r_mean;
  double *p_g, r_g;
  double *pro, *prs, *pr;
  double *pb, *pbb, diff;
  double *pbb_over;
  double *pgl;

  int drg = *dr * *dg;
  double *pr_smooth_h, *pr_scaled_h;
  double *pr_sc, *pr_sc_end;
  int span, s, i_rand;
  double *pe;
  double *pa;
  double *prh, *prh_end;

  double *e_matrix_end;
  double *a_random_end;
  e_matrix_end = e_matrix + drg;
  a_random_end = a_random + drg;

  for (i = 0, pb = b; i < *n; ++i) {  /* foreach ROUND */
    if (!(i % 100)) {
      printf("%d ",i); fflush(0);
    }

    pa = a_random;
    pe = e_matrix;
    while (pa < a_random_end)
      *(pa++) = *(pe++);

    /* random shuffle r and add to a_random */
    for (h = 0; h < *dh; ++h) {
      pr_smooth_h = r_smooth + drg*h;
      pr_scaled_h = r_scaled + drg*h;
      pa = a_random;
      span = spans[h];
      for (s = 0; s < drg/span; ++s) {
        i_rand = (int)(my_rand() * drg/span);
        /* rand = s; */
        pr_sc = pr_scaled_h + i_rand*span;
        pr_sc_end = pr_sc + span;
        for ( ; pr_sc < pr_sc_end;
            ++pr_sc, ++pr_smooth_h, ++pa) {
          *pa += *pr_smooth_h * *pr_sc;
        }
      }
    }

    /* average G over all genes and rest */
    r_mean = 0;
    for (p1 = a_random; p1 < a_random_end; ++p1)
      r_mean += *p1;
    r_mean /= drg; /* close to 0 anyway */

    /* gene average and comparison */ 
    gg = 0;
    if (*stats_only) {
      pb = b;
      pbb = b + *dg;
      pbb_over = b + 2*(*dg);
      /* p_g = g; */
      for (pg1 = a_random, j = 0, pgl = gene_limit;  /* foreach GENE */
          pg1 < a_random_end;
          pg1 += *dr, ++j, ++pgl,
          ++pb, ++pbb, ++pbb_over) {
        /* printf("  g %d\n",gg++); */
        g_end = pg1 + *dr;
        r_g = 0;
        for (p1 = pg1; p1 < g_end; ++p1)
          r_g += *p1;
        r_g /= *dr;
        /* gene effect is gene average - total mean mu */
        diff = r_g - r_mean;
        *pb += diff;
        *pbb += diff*diff;
        /* if (j == 3376) {
           printf(" %lf ",diff); fflush(0);
        } */
        if (diff > *pgl)
          (*pbb_over)++;
      }
    }
    else
      for (pg1 = a_random; pg1 < a_random_end;
          pg1 += *dr, ++pb) {
        /* printf("  g %d\n",gg++); */
        g_end = pg1 + *dr;
        r_g = 0;
        for (p1 = pg1; p1 < g_end; ++p1)
          r_g += *p1;
        r_g /= *dr;
        *pb = r_g - r_mean;
        /* *pb = r_g; */
      }
  }
}



/**
 * Routine for fast normal equation coefficients
 */

/* R stores in column major, so columns first! */

void norm_coefs(
    /* as.integer(m), */
    int* m,
    /* as.double(m.inters), */
    /*  as.integer(nrow(m.inters)),as.integer(ncol(m.inters)), */
    int* tm_inters, int* r_tm_inters, int* c_tm_inters,
    /* as.integer(na.m), */
    /*   as.integer(nrow(na.m)),as.integer(ncol(na.m)), */
    int* t_na_m, int* r_t_na_m, int* c_t_na_m,
    /* as.double(prod.dim.inters), */
    /*   as.integer(length(prod.dim.inters)), */
    double* prod_dim_inters, int* n_prod_dim_inters,
    /* as.double(prod.dims), */
    double* prod_dims,
    /* as.logical(is.sub.inters), */
    /*  as.integer(nrow(is.sub.inters)),as.integer(ncol(is.sub.inters)), */
    int* is_sub_inters, int* r_is_sub_inters, int* c_is_sub_inters,
    /* as.double(est.coefs), */
    /*  as.integer(length(est.coefs)) */
    double* est_coefs, int* n_est_coefs,
    /* norm.coefs=as.double(norm.coefs) */
    /*  as.integer(nrow(norm,coefs)), */
    /*  as.integer(ncol(norm,coefs))  */
    double* norm_coefs, int* r_norm_coefs, int* c_norm_coefs
  ) {

  int e, y, i, j, d, l, sub_i;
  int* t_na_m_y; int* t_na_m_e;
  int* inters_i;
  int* is_sub_inters_i;
  /* double* est_coefs; */
  double sum_coefs;
  double c;
  /* est_coefs = (double*)calloc(*m,sizeof(double)); */

  /* Did everything arrive here */
  /*
  for (i = 0; i < *n_prod_dim_inters; ++i)
    printf("%e ",prod_dim_inters[i]);
  printf("\n");
  printf("prod dims %e \n",*prod_dims);

  for (e = 0; e < *m; ++e) {
    printf("e %d \n",e);
    for (y = 0; y < *m; ++y) {
      printf(" y %d \n",y);
        t_na_m_y = &t_na_m[*r_t_na_m * y];
        t_na_m_e = &t_na_m[*r_t_na_m * e];
        for (d = 0; d < *r_tm_inters; ++d) {
          printf("    %d %d\n",t_na_m_y[d],t_na_m_e[d]);
        }
    }
  }

  for (i = 0; i < *c_tm_inters; ++i) {
        inters_i = &tm_inters[*r_tm_inters * i];
        for (d = 0; d < *r_tm_inters; ++d) {
          printf("%d ",inters_i[d]);
        }
    printf("\n");
  }
  */

  for (e = 0; e < *m; ++e) {
    printf("%d ",e);
    for (y = 0; y < *m; ++y) {
      /* printf(" y %d \n",y); */
      for (l = 0; l < *n_est_coefs; ++l)
        est_coefs[l] = 0;
      sum_coefs = 0;
      for (i = 0; i < *c_tm_inters; ++i) {
        /* printf("   i %d \n",i); */
        c = -*prod_dims;
        t_na_m_y = &t_na_m[*r_t_na_m * y];
        t_na_m_e = &t_na_m[*r_t_na_m * e];
        inters_i = &tm_inters[*r_tm_inters * i];
        for (d = 0; d < *r_tm_inters; ++d) {
          if (inters_i[d] && t_na_m_y[d] != t_na_m_e[d])
            break;
        }
        if (d == *r_tm_inters)
          c += prod_dim_inters[i];
        is_sub_inters_i = &is_sub_inters[*r_is_sub_inters * i];
        for (sub_i = 0; sub_i < i; ++sub_i)
          if (is_sub_inters_i[sub_i])
            c -= est_coefs[sub_i];
        sum_coefs += est_coefs[i] = c;
      }
      /* printf("    ");
      for (l = 0; l < *n_est_coefs; ++l)
        printf("%e ",est_coefs[l]);
      printf("\n sum %e  total %e\n",sum_coefs,*prod_dims + sum_coefs); */
      norm_coefs[*r_norm_coefs * y + e] = *prod_dims + sum_coefs;
    }
  }
}

/*
 * SAM stuff 
 */

int compare_double(const double* sa1, const double* sa2) {
  return (double)*sa1 > (double)*sa2;
}

void sam_two_class(
    int *cycles,
    int *perms,   /* int matrix: (m1 + m2) x nperms */
    int *nperms,
    int *do_row_perms,
    int *n,
    int *m1,
    int *m2,
    double *s0,
    double *tf,   /* double matrix: (m1 + m2) x n */
    double *dall  /* double matrix: n x cycles */
    ) {

  int round;
  int m = *m1 + *m2;
  double *tf_end = tf + m*(*n);
  double *pf;
  double mean1, mean2;
  int *pind;
  double *pdall, *pdall_sort;
  int *pperm, *pperm2, *ind_end;
  int *pperm_end = perms + m*(*nperms);
  double t, x1,s1,x2,s2;
  double si;
  double s_factor;
  s_factor = sqrt((1.0/(double)*m1 + 1.0/(double)*m2)/(*m1 + *m2 - 2.0));

  for (round = 0, pperm = perms, pdall_sort = dall;
      round < *cycles;
      pperm += m, pdall_sort += *n, ++round) {
    /* printf("round %d \n",round); fflush(0); */
    printf("%d ",round); fflush(0);

    pperm2 = pperm + *m1;
    ind_end = pperm2 + *m2;
    for (pf = tf, pdall = pdall_sort;
        pf < tf_end;
        pf += m, ++pdall) {
      /* for each of n genes */
      mean1 = mean2 = 0;
      s1 = s2 = 0;

      if (*do_row_perms) {
        pperm = perms + ((int)(my_rand() * *nperms))*m;
        pperm2 = pperm + *m1;
        ind_end = pperm2 + *m2;
      }
      for (pind = pperm; pind < pperm2; ++pind) {
        /* for each of m experiments */
        x1 = *(pf + *pind);
        /* printf("%d: %lf ",*pind,x1); fflush(0); */
        mean1 += x1;
        s1 += x1*x1;
      }
      mean1 /= *m1;
      for (pind = pperm2; pind < ind_end; ++pind) {
        /* for each of m experiments */
        x2 = *(pf + *pind);
        mean2 += x2;
        s2 += x2*x2;
      }
      mean2 /= *m2;
      si = s_factor*sqrt((s1 - mean1*mean1*(*m1)) +
          (s2 - mean2*mean2*(*m2)));
      /* printf("m1 %lf m2 %lf si %lf\n",mean1,mean2,si); fflush(0); */
      *pdall = (mean1 - mean2) / (si + *s0);
    }
    qsort(pdall_sort,*n,sizeof(double),compare_double);
  }
  printf("\n");
}

void sam_single(
    int *cycles,
    int *signs,   /* int matrix: (m) x nsigns */
    int *nsigns,
    int *do_row_signs,
    int *n,
    int *m,
    double *s0,
    double *tf,   /* double matrix: (m) x n */
    double *dall  /* double matrix: n x cycles */
    ) {

  int round;
  double *tf_end = tf + *m*(*n);
  double *pf;
  double mean;
  double *pval, *val_end;
  double *pdall, *pdall_sort;
  int *psign, *pind_sign;
  double s1;
  double si;
  double s_factor;
  s_factor = sqrt(1.0/(double)(*m)/(double)(*m-1));

  for (round = 0, psign = signs, pdall_sort = dall;
      round < *cycles;
      psign += *m, pdall_sort += *n, ++round) {
    printf("%d ",round); fflush(0);

    for (pf = tf, pdall = pdall_sort;
        pf < tf_end;
        pf += *m, ++pdall) {
      /* for each of n genes */
      mean  = 0;
      s1 = 0;

      if (*do_row_signs) {
        psign = signs + ((int)(my_rand() * *nsigns))*(*m);
      }
      val_end = pf + *m;
      for (pval = pf, pind_sign = psign;
          pval < val_end;
          ++pval, ++pind_sign) {
        /* for each of m experiments */
        mean += (*pval)*(*pind_sign);
        s1 += (*pval)*(*pval);
      }
      mean /= *m;
      si = s_factor*sqrt(s1 - mean*mean*(*m));
      *pdall = mean / (si + *s0);
    }
    qsort(pdall_sort,*n,sizeof(double),compare_double);
  }
  printf("\n");
}


// P. S. patch - add dynamical load information.
/* 
void t_boots(
    int *cycles,
    int *n,
    int *m,
    int *step_down,
    double *s,
    double *tf,
    int *ind,
    int *count)

void vg_boots(
    int *n,    
    int *dr, int *dg, int *dv,  
    double *e,  
    double *dvg,
    double *r_smooth, double *r_scaled,
    double *r,  
    int *stats_only,
    double *b  
  )
void g_boots(
    int *n,    
    int *dr,
    int *dg,   
    double *e, 
    double *r_smooth,
    double *r_scaled,
    double *r,  
    int *stats_only,
    double *b  
  )
void g_level_boots(
    int *n,    
    int *dh,   
    int *dr,
    int *dg,   
    double *e_matrix,
    double *a_random,
    double *r_smooth,
    double *r_scaled,
    double *gene_limit,
    int *spans,  
    int *stats_only,
    double *b 
  )
void norm_coefs(
    int* m,
    int* tm_inters, int* r_tm_inters, int* c_tm_inters,
    int* t_na_m, int* r_t_na_m, int* c_t_na_m,
    double* prod_dim_inters, int* n_prod_dim_inters,
    double* prod_dims,
    int* is_sub_inters, int* r_is_sub_inters, int* c_is_sub_inters,
    double* est_coefs, int* n_est_coefs,
    double* norm_coefs, int* r_norm_coefs, int* c_norm_coefs
  )
void sam_two_class(
    int *cycles,
    int *perms, 
    int *nperms,
    int *do_row_perms,
    int *n,
    int *m1,
    int *m2,
    double *s0,
    double *tf, 
    double *dall
    )
void sam_single(
    int *cycles,
    int *signs, 
    int *nsigns,
    int *do_row_signs,
    int *n,
    int *m,
    double *s0,
    double *tf, 
    double *dall
    ) 
numeric  REALSXP 
integer  INTSXP 
logical  LGLSXP 
single  SINGLESXP 
character  STRSXP 
list  VECSXP
*/ 

R_CMethodDef cMethods[] = {
  {"rand_interface", &rand_interface, 1, {REALSXP}},   // just for PS tests!
  {"t_boots", &t_boots, 8, {INTSXP, INTSXP, INTSXP, INTSXP, REALSXP, REALSXP, INTSXP, INTSXP}},
  {"vg_boots", &vg_boots, 11, {INTSXP, INTSXP, INTSXP, INTSXP, REALSXP, REALSXP, REALSXP, 
			       REALSXP, REALSXP, INTSXP, REALSXP}},
  {"g_boots", &g_boots, 9, {INTSXP, INTSXP, INTSXP, REALSXP, REALSXP, REALSXP, 
			    REALSXP, INTSXP, REALSXP}},
  {"g_level_boots", &g_level_boots, 12, {INTSXP, INTSXP, INTSXP, INTSXP, REALSXP, REALSXP, REALSXP, 
					 REALSXP, REALSXP, INTSXP, INTSXP,REALSXP}},
  {"norm_coefs", &norm_coefs, 18, {INTSXP, INTSXP, INTSXP, INTSXP, INTSXP, INTSXP, INTSXP, 
				   REALSXP, INTSXP, REALSXP, INTSXP, INTSXP, INTSXP, REALSXP,
				   INTSXP, REALSXP, INTSXP, INTSXP}},
  {"sam_two_class", &sam_two_class, 10, {INTSXP, INTSXP, INTSXP, INTSXP, INTSXP, INTSXP, INTSXP, 
					 REALSXP, REALSXP, REALSXP}},
  {"sam_single", &sam_single, 9, {INTSXP, INTSXP, INTSXP, INTSXP, INTSXP, INTSXP,
					 REALSXP, REALSXP, REALSXP}},
  {NULL, NULL, 0}
};

void
R_init_mylib(DllInfo *info)
{
  /* Register routines, allocate resources. */
  R_registerRoutines(info, cMethods, NULL, NULL, NULL);
#ifdef __MINGW32__
  GetRNGstate();
#endif
}

void
R_unload_mylib(DllInfo *info)
{
  /* Release resources. */
#ifdef __MINGW32__
  PutRNGstate();
#endif
}
