
#/**
# @name
# \title{t-statistic bootstrap for differential gene expression}
# \description{
#   Calculates p-values for over- or under-expression from bootstrapping
#   a t-statistic for each gene's \eqn{\log(R/G)}{log(R/G)}
# }
# @usage
# @synopsis
# \arguments{
#   \item{RG}{RG microarray data structure, see \code{\link{read.rg}}}
#   \item{rounds}{bootstrap resamples}
#   \item{fast}{if T, the compiled bootstrap is called which is
#     \emph{much} faster then the R version, but might not be
#     available on every platform }
#   \item{step.down}{if T, the Westfall-Young step-down adjustment for
#       multiple testing is used otherwise no adjustement is
#       applied. Alternatively, if set to F the Benjamini-Hochberg's
#       FDR \code{\link{pvalue.fdr}} adjustment might be applied to
#       the result}
#   \item{centralize}{if TRUE, shifts all experiments before comparison}
# }
# \value{
#   List containing:
#   \item{p.over, p.under}{p-values of over-, under-expressed genes}
#   \item{rounds}{bootstrap resamples}
#   \item{RG}{the input RG microarray data structure}
# }
# \examples{
#   data(trcs)
#   RG <- rg.remove.quantile(trcs.RG,0.1,level=1)
#   ou <- t.boots(RG,1000)
#   ldf <- data.over.under(ou,pvalue=0.05,method="none")
#   ldf$over
#   ou$p.over
# }
# \seealso{
#   \code{\link{t.statistic}},
#   \code{\link{data.over.under}}}
# }
# \keyword{htest}
#*/

t.boots <- function(RG,rounds=10000,fast=T,step.down=T,centralize=T) {
  list(p.over=t.boots.over(RG,rounds,over=T,fast,step.down,centralize=T),
    p.under=t.boots.over(RG,rounds,over=F,fast,step.down,centralize=T),
    rounds=rounds,
    RG=RG)
}

t.boots.over <- function(RG,rounds=10000,over=T,fast=T,
    step.down=T,centralize=T) {
  f <- rg.log.matrix(RG)
  if (!over)
    f <- -f
  if (centralize)
    f <- scale(f,scale=F)
  n <- nrow(f)
  m <- ncol(f)

  avg <- apply(f,1,mean)
  f.null <- f-avg
  m.sd <- sqrt((m-1)*apply(f,1,var))
  # variance: m.sd * sqrt(1/m)*sqrt(1/(m-1))
  #ts <- abs(avg/m.sd)
  ts <- avg/m.sd
  #ts <- avg

  o <- order(ts)
  to <- ts[o]
  fno <- f.null[o,]
  count <- rep(0,n)

  if (fast) {
    ind <- rep(0,m)
    count <- .C("t_boots",
      as.integer(rounds),
      as.integer(n),
      as.integer(m),
      as.integer(step.down),
      as.double(to),
      as.double(t(fno)),
      as.integer(ind),
      count=as.integer(count)
    )$count
  } else {
    for (rnd in 1:rounds) {
      cat(rnd," ")
      ind.s <- sample(1:m,replace=T)
      avg.s <- apply(fno[,ind.s],1,mean)
      m.sd.s <- sqrt((m-1)*apply(fno[,ind.s],1,var))
      tvec <- avg.s/m.sd.s
      #tvec <- avg.s

      # min step-down for p is max step-up for t:
      if (step.down) {
        max.i <- tvec[1]
        for (i in 2:n) {
          if (tvec[i] > max.i) 
            max.i <- tvec[i]
          if (max.i > to[i])
            count[i] <- count[i] + 1
        }
      } else {
        tvec > to
      }
    }
    cat("\n")
  }

  p <- count/rounds

  # order increasingly (small p = high t)
  for (i in (n-1):1) {
    p[i] <- max(p[i],p[i+1])
  }

  p.genes <- rep(0,n)
  p.genes[o] <- p
  p.genes
}

#/**
# \name{t.statistic}
# \alias{t.statistic}
# \alias{t.statistic.2}
# \title{t-statistic for differential gene expression}
# \description{
#   Calculates p-values for over- or under-expression using t-statistics
#   on log(R/G) values.
# }
# \usage {
#   t.statistic(RG,centralize=T)
#   t.statistic.2(RG,m=NULL,centralize=T)
# }
# @synopsis
# \arguments{
#   \item{RG}{RG microarray data structure, see \code{\link{read.rg}} }
#   \item{centralize}{if TRUE, shifts all experiments before comparison }
#   \item{m}{the number of experiments for the first mean tested
#     against the the mean of the rest of experiments}
# }
# \details{
#   Performs a t-test on log ratio values of each gene, either to test
#   against 0 (\code{t.statistic}) or to test the \code{m} first
#   experiments against the rest.
# }
# \value{
#   List containing:
#   \item{p.over, p.under}{p-values of over-, under-expressed genes}
#   \item{RG}{the input RG microarray data structure}
# }
# \examples{
#   data(trcs)
#   RG <- rg.remove.quantile(trcs.RG,0.1,level=1)
#   ou <- t.statistic(RG)
#   ldf <- data.over.under(ou,pvalue=0.05,method="bonferroni")
#   ldf$over
#   ou$p.over
# }
# \seealso{
#   \code{\link{t.boots}},
#   \code{\link{data.over.under}}
# }
# \keyword{htest}
#*/

t.statistic <- function(RG,centralize=T) {
  f <- rg.log.matrix(RG)
  list(p.over=t.statistic.over(f,over=T,centralize),
    p.under=t.statistic.over(f,over=F,centralize),
    RG=RG)
}

t.statistic.2 <- function(RG,centralize=T,m1=NULL) {
  f <- rg.log.matrix(RG)
  list(p.over=t.statistic.2.over(f,over=T,m1=m1,centralize),
    p.under=t.statistic.2.over(f,over=F,m1=m1,centralize),
    RG=RG)
}

t.statistic.over <- function(f,over=T,centralize=T) {
  if (centralize)
    f <- scale(f,scale=F)
  if (!over)
    f <- -f
  n <- nrow(f)
  m <- ncol(f)

  avg <- apply(f,1,mean)
  sd <- apply(f,1,sd)
  tv <- avg/(sd/sqrt(m))
  pt(tv,m-1,lower.tail=F)
}

t.statistic.2.over <- function(f,over=T,m1=NULL,centralize=T) {
  if (centralize)
    f <- scale(f,scale=F)
  n <- nrow(f)
  m <- ncol(f)
  if (is.null(m1))
    m1 <- m/2

  avg1 <- apply(f[,1:m1],1,mean)
  avg2 <- apply(f[,(m1+1):m],1,mean)
  var1 <- apply(f[,1:m1],1,var)
  var2 <- apply(f[,(m1+1):m],1,var)
  sd <- sqrt(((m1-1)*var1 + (m-m1-1)*var2)/(m-2))
  if (over)
    davg <- avg1 - avg2
  else
    davg <- avg2 - avg1
  tv <- davg/(sd*sqrt(1/m1 + 1/(m-m1)))
  pt(tv,m-2,lower.tail=F)
}




